#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2021 Oleksandr Moskalenko <om@rc.ufl.edu>
#
# Distributed under terms of the MIT license.

"""
Visualize fragments and events in a terminal
"""

def print_region(data):
    """
    Print a scaled region to screen with labels and dashes for visualization.
    """
    label = ""
    region = ""
    r_end = 0
    for i in (sorted(data.keys())):
        start = i
        name = data[start][0]
        end = data[start][1]
        ilen = end - start - 1
        if ilen < 0:
            ilen = 0
        gap = start - r_end - 1
        if gap <= 0:
            gap = 0
        region += gap * " " + '|'
        label += (gap + 1) * " "
        region += (ilen * "-") + '|'
        r_end = end + 1
        l_pad = round((end - start) / 2 - len(name) / 2) - 1
        if l_pad < 0:
            l_pad = 0
        label += l_pad * " "
        label += name
        # print(f"ER: {name}, {start}, {end}, {ilen}, {r_end}, {gap}, {l_pad}")
        # print(label)
        # print(region)
        l_fill = len(region) - len(label)
        label += l_fill * " "
    print(label)
    print(region)


def plot_ers(ers, tx1, tx2):
    """
    Plot ERs visually when debugging.
    """
    start, end, er_dict, tx1_dict, tx2_dict = [], [], {}, {}, {}
    for er in ers:
        start.append(er.start)
        end.append(er.end)
    g_start = min(start)
    g_end = max(end)
    er_range = g_end - g_start + 1
    scale = 200.0 / er_range
    print(tx1_dict)
    for i in ers:
        i_name = i.name.split(':')[1]
        er_dict[round((i.start - g_start) * scale)] = [i_name, round((i.end - g_start) * scale)]
    for i in tx1:
        tx1_name = i.name.split("_")[0]
        i_name = "E" + i.name.split('_')[-1]
        tx1_dict[round((i.start - g_start) * scale)] = [i_name, round((i.end - g_start) * scale)]
    for i in tx2:
        tx2_name = i.name.split("_")[0]
        i_name = "E" + i.name.split('_')[-1]
        tx2_dict[round((i.start - g_start) * scale)] = [i_name, round((i.end - g_start) * scale)]
    gene_name = next(iter(er_dict.values()))[0].split(':')[0]
    print(f"Exonic Regions for {gene_name} gene")
    print('|' + '-' * 198 + '|')
    print_region(er_dict)
    print('|' + '-' * 98 + '|')
    print(f"Transcript 1: {tx1_name}")
    print_region(tx1_dict)
    print('|' + '-' * 98 + '|')
    print(f"Transcript 2: {tx2_name}")
    print_region(tx2_dict)
    print('|' + '-' * 98 + '|')






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 13 15:03:09 2021

@author: adalena.nanni
"""

import pandas as pd
import numpy as np


td_df_cols = [
    "gene_id",
    "transcript_1",
    "transcript_2",
    "num_junction_T1_only",
    "num_junction_T2_only",
    "num_junction_shared",
    "prop_junction_diff",
    "prop_junction_similar",
    "junction_T1_all",
    "junction_T2_all",
    "junction_T1_only",
    "junction_T2_only",
    "junction_shared",
    "num_ER_T1_only",
    "num_ER_T2_only",
    "num_ER_shared",
    "prop_ER_diff",
    "prop_ER_similar",
    "ER_T1_only",
    "ER_T2_only",
    "ER_shared",
    "num_fragment_T1_only",
    "num_fragment_T2_only",
    "num_fragment_shared",
    "prop_fragment_diff",
    "prop_fragment_similar",
    "fragment_T1_only",
    "fragment_T2_only",
    "fragment_shared",
    "num_fragment_singletons_T1_only",
    "num_fragment_singletons_T2_only",
    "num_fragment_singletons_shared",
    "num_IR_fragment_T1",
    "num_IR_fragment_T2",
    "IR_fragment_T1",
    "IR_fragment_T2",
    "num_nt_shared",
    "num_nt_T1_only",
    "num_nt_T2_only",
    "num_nt_diff",
    "total_nt",
    "prop_nt_diff",
    "prop_nt_similar",
    "num_nt_T1_only_in_shared_ER",
    "num_nt_T2_only_in_shared_ER",
    "num_nt_shared_in_shared_ER",
    "total_nt_in_shared_ER",
    "prop_nt_diff_in_shared_ER",
    "prop_nt_similar_in_shared_ER",
    "num_nt_T1_only_in_unique_ER",
    "num_nt_T2_only_in_unique_ER",
    "flag_FSM",
    "flag_T1_ISM_of_T2",
    "flag_T2_ISM_of_T1",
    "flag_IR",
    "flag_5_variation",
    "flag_3_variation",
    "flag_alt_donor_acceptor",
    "flag_alt_exon",
    "flag_no_shared_nt",
]


def calculate_distance(out_df, junction_df, gene_id, tx1_name, tx2_name, fsm=False):
    # Check for monoexons
    has_junctions = True
    if junction_df.empty:
        has_junctions = False
    # Set EF variable
    ef_df = out_df.copy()

    # Make empty Series to place distance values
    singlePair = pd.Series(index=td_df_cols, dtype=object)
    singlePair[["gene_id", "transcript_1", "transcript_2"]] = (
        gene_id,
        tx1_name,
        tx2_name,
    )

    # Set exon fragment and exon region ID values based on genomic coordinates
    ef_df["fragment_id"] = (
        ef_df["ef_chr"].map(str)
        + ":"
        + ef_df["ef_start"].map(str)
        + ":"
        + ef_df["ef_end"].map(str)
        + ":"
        + ef_df["ef_strand"].map(str)
    )
    ef_df["region_id"] = (
        ef_df["er_chr"].map(str)
        + ":"
        + ef_df["er_start"].map(str)
        + ":"
        + ef_df["er_end"].map(str)
        + ":"
        + ef_df["er_strand"].map(str)
    )

    # Flag shared fragments
    ef_df["flag_shared"] = np.where(
        (ef_df["transcript_id"] == tx1_name + "|" + tx2_name)
        | (ef_df["transcript_id"] == tx2_name + "|" + tx1_name),
        1,
        0,
    )

    # Flag singleton fragments (where the fragment is a full exon region)
    ef_df["num_fragment_in_ER"] = ef_df.groupby("er_id")["ef_id"].transform("count")
    ef_df["flag_singleton"] = np.where(ef_df["num_fragment_in_ER"] == 1, 1, 0)

    # Add fragment lengths (end - start)
    ef_df["fragment_length"] = ef_df["ef_end"].map(int) - ef_df["ef_start"].map(int)

    # Sort junctions by chromosome and coordinates
    if has_junctions:
        sorted_junction_df = junction_df.copy()
        sorted_junction_df[["chr", "start", "end", "strand"]] = (
                sorted_junction_df["coords"].str.split(":", expand=True)
        )
        sorted_junction_df[["start", "end"]] = (
                sorted_junction_df[["start", "end"]].astype(int)
        )
        sorted_junction_df = sorted_junction_df.sort_values(
                by=["chr", "start", "end"]
            ).reset_index(drop=True)
    else:
        sorted_junction_df = pd.DataFrame()

    # Get distance measures for junctions
    singlePair = get_junction_distance(
        singlePair, sorted_junction_df, tx1_name, tx2_name, fsm=fsm
    )

    # Gt distance measures for exon regions (ER) and exon fragments (EF)
    singlePair, ERSharedSet = get_ER_distance(
        singlePair, ef_df, tx1_name, tx2_name, fsm=fsm
    )
    singlePair = get_EF_distance(singlePair, ef_df, tx1_name, tx2_name, ERSharedSet)

    # Set flags for different alternative splicing (AS) events
    singlePair = set_AS_flags(singlePair, ef_df, tx1_name, tx2_name, has_junctions)

    # Return distance of transcript pair
    return singlePair


def get_junction_distance(singlePair, sorted_junction_df, tx1_name, tx2_name, fsm=False):
    # Check if transcript pair shares all junctions (FSM, full-splice match)
    if fsm:
        # Check if both transcripts are monoexon (no junctions)
        if len(sorted_junction_df) == 0:
            singlePair[
                ["num_junction_T1_only", "num_junction_T2_only", "num_junction_shared"]
            ] = 0
            singlePair["prop_junction_diff"] = 0
            singlePair["prop_junction_similar"] = 1
            singlePair[
                ["junction_T1_only",
                 "junction_T2_only",
                 "junction_shared",
                 "junction_T1_all",
                 "junction_T2_all"]
            ] = ""
        # FSM transcripts are multiexon
        else:
            singlePair[["num_junction_T1_only", "num_junction_T2_only"]] = 0
            singlePair["num_junction_shared"] = len(
                sorted_junction_df[sorted_junction_df["transcript_id"] == tx1_name]
            )
            singlePair["prop_junction_diff"] = 0
            singlePair["prop_junction_similar"] = 1
            singlePair[["junction_T1_only", "junction_T2_only"]] = ""
            singlePair[
                    ["junction_shared", "junction_T1_all", "junction_T2_all"]
                ] = "|".join(
                    sorted_junction_df[
                            sorted_junction_df["transcript_id"] == tx1_name
                        ]["coords"]
            )

    # Transcript pair does not share all junctions
    else:
        # Check if both transcripts are monoexon but do not overlap (not fsm)
        if len(sorted_junction_df) == 0:
            singlePair[
                ["num_junction_T1_only", "num_junction_T2_only", "num_junction_shared"]
            ] = 0
            singlePair["prop_junction_diff"] = 0
            singlePair["prop_junction_similar"] = 1
            singlePair[
                ["junction_T1_only",
                 "junction_T2_only",
                 "junction_shared",
                 "junction_T1_all",
                 "junction_T2_all"]
            ] = ""
        # Check if one of the transcripts is monoexon (only junctions from one transcript and not
        # the other)
        elif sorted_junction_df["transcript_id"].nunique() == 1:
            # Only T1 is monoexon
            if tx1_name in sorted_junction_df["transcript_id"].unique():
                singlePair[["num_junction_T1_only", "num_junction_shared"]] = 0
                singlePair["prop_junction_similar"] = 0
                singlePair["prop_junction_diff"] = 1
                singlePair[
                    ["junction_T1_only", "junction_T1_all", "junction_shared"]
                ] = ""
                singlePair["num_junction_T2_only"] = len(sorted_junction_df)
                singlePair[
                    ["junction_T2_only", "junction_T2_all"]
                ] = "|".join(sorted_junction_df["coords"])
            # Only T2 is monoexon
            else:
                singlePair[["num_junction_T2_only", "num_junction_shared"]] = 0
                singlePair["prop_junction_similar"] = 0
                singlePair["prop_junction_diff"] = 1
                singlePair[
                    ["junction_T2_only", "junction_T2_all", "junction_shared"]
                ] = ""
                singlePair["num_junction_T1_only"] = len(sorted_junction_df)
                singlePair[
                    ["junction_T1_only", "junction_T1_all"]
                ] = "|".join(sorted_junction_df["coords"])
        # Both transcripts multi-exon
        else:
            t1Junc = sorted_junction_df[
                    sorted_junction_df["transcript_id"] == tx1_name
                ]["coords"]
            t2Junc = sorted_junction_df[
                    sorted_junction_df["transcript_id"] == tx2_name
                ]["coords"]
            singlePair["junction_T1_all"] = "|".join(t1Junc)
            singlePair["junction_T2_all"] = "|".join(t2Junc)
            juncSharedSet = set(t1Junc).intersection(set(t2Junc))
            juncT1Set = set(t1Junc).difference(set(t2Junc))
            juncT2Set = set(t2Junc).difference(set(t1Junc))
            singlePair["num_junction_T1_only"] = len(juncT1Set)
            singlePair["num_junction_T2_only"] = len(juncT2Set)
            singlePair["num_junction_shared"] = len(juncSharedSet)
            singlePair["prop_junction_diff"] = (
                singlePair["num_junction_T1_only"] + singlePair["num_junction_T2_only"]
            ) / (
                singlePair["num_junction_T1_only"]
                + singlePair["num_junction_T2_only"]
                + singlePair["num_junction_shared"]
            )
            singlePair["prop_junction_similar"] = 1 - singlePair["prop_junction_diff"]
            singlePair["junction_T1_only"] = "|".join(
                sorted(juncT1Set, key=lambda x: t1Junc[t1Junc == x].index[0])
            )
            singlePair["junction_T2_only"] = "|".join(
                sorted(juncT2Set, key=lambda x: t2Junc[t2Junc == x].index[0])
            )
            singlePair["junction_shared"] = "|".join(
                sorted(juncSharedSet, key=lambda x: t2Junc[t2Junc == x].index[0])
            )
    return singlePair


def get_ER_distance(singlePair, ef_df, tx1_name, tx2_name, fsm=False):
    # Check if transcript pair shares all junctions (FSM, full-splice match)
    if fsm:
        # All junctions are shared, therefore all ER are shared
        singlePair[["num_ER_T1_only", "num_ER_T2_only"]] = 0
        singlePair["num_ER_shared"] = ef_df["er_id"].nunique()
        singlePair["prop_ER_diff"] = 0
        singlePair["prop_ER_similar"] = 1
        singlePair[["ER_T1_only", "ER_T2_only"]] = ""
        singlePair["ER_shared"] = "|".join(ef_df["region_id"].unique())
        ERSharedSet = ef_df["region_id"].drop_duplicates()
    else:
        # Check for shared and unique ER
        ERall = (
            ef_df.groupby("region_id")
            .agg({"transcript_id": (lambda x: max(x, key=len)), "flag_shared": "max"})
            .reset_index()
        )
        ERSharedSet = ERall[ERall["flag_shared"] == 1]["region_id"]
        ERT1Set = ERall[
            (ERall["flag_shared"] == 0) & (ERall["transcript_id"] == tx1_name)
        ]["region_id"]
        ERT2Set = ERall[
            (ERall["flag_shared"] == 0) & (ERall["transcript_id"] == tx2_name)
        ]["region_id"]
        singlePair["num_ER_T1_only"] = len(ERT1Set)
        singlePair["num_ER_T2_only"] = len(ERT2Set)
        singlePair["num_ER_shared"] = len(ERSharedSet)
        singlePair["prop_ER_diff"] = (
            singlePair["num_ER_T1_only"] + singlePair["num_ER_T2_only"]
        ) / (
            singlePair["num_ER_T1_only"]
            + singlePair["num_ER_T2_only"]
            + singlePair["num_ER_shared"]
        )
        singlePair["prop_ER_similar"] = 1 - singlePair["prop_ER_diff"]
        singlePair["ER_T1_only"] = "|".join(ERT1Set)
        singlePair["ER_T2_only"] = "|".join(ERT2Set)
        singlePair["ER_shared"] = "|".join(ERSharedSet)
    return singlePair, ERSharedSet


def get_EF_distance(singlePair, ef_df, tx1_name, tx2_name, ERSharedSet):
    num_shared = ef_df["flag_shared"].sum()
    # Check if all fragments are shared (transcripts are identical)
    if num_shared == len(ef_df):
        # All EF shared, transcripts are identical
        singlePair[["num_fragment_T1_only", "num_fragment_T2_only"]] = 0
        singlePair["num_fragment_shared"] = len(ef_df)
        singlePair["prop_fragment_diff"] = 0
        singlePair["prop_fragment_similar"] = 1
        singlePair[["fragment_T1_only", "fragment_T2_only"]] = ""
        singlePair["fragment_shared"] = "|".join(ef_df["fragment_id"])
        singlePair[
            ["num_fragment_singletons_T1_only", "num_fragment_singletons_T2_only"]
        ] = 0
        singlePair["num_fragment_singletons_shared"] = singlePair["num_fragment_shared"]

        # Nucleotide differences are 0, all are shared and are in shared ER
        singlePair[
            [
                "total_nt",
                "total_nt_in_shared_ER",
                "num_nt_shared",
                "num_nt_shared_in_shared_ER",
            ]
        ] = ef_df["fragment_length"].sum()
        singlePair[
            [
                "num_nt_T1_only",
                "num_nt_T1_only_in_shared_ER",
                "num_nt_T1_only_in_unique_ER",
                "num_nt_T2_only",
                "num_nt_T2_only_in_shared_ER",
                "num_nt_T2_only_in_unique_ER",
                "num_nt_diff",
            ]
        ] = 0
        singlePair[["prop_nt_diff", "prop_nt_diff_in_shared_ER"]] = 0
        singlePair[["prop_nt_similar", "prop_nt_similar_in_shared_ER"]] = 1

        # No IR present between identical transcripts
        singlePair[["num_IR_fragment_T1", "num_IR_fragment_T2"]] = 0
        singlePair[["IR_fragment_T1", "IR_fragment_T2"]] = ""
    else:
        # Not all EF shared
        fragSharedSet = ef_df[ef_df["flag_shared"] == 1]
        fragSharedSingSet = ef_df[
            (ef_df["flag_shared"] == 1) & (ef_df["flag_singleton"] == 1)
        ]
        fragT1Set = ef_df[
            (ef_df["flag_shared"] == 0) & (ef_df["transcript_id"] == tx1_name)
        ]
        fragT1SingSet = ef_df[
            (ef_df["flag_shared"] == 0)
            & (ef_df["transcript_id"] == tx1_name)
            & (ef_df["flag_singleton"] == 1)
        ]
        fragT2Set = ef_df[
            (ef_df["flag_shared"] == 0) & (ef_df["transcript_id"] == tx2_name)
        ]
        fragT2SingSet = ef_df[
            (ef_df["flag_shared"] == 0)
            & (ef_df["transcript_id"] == tx2_name)
            & (ef_df["flag_singleton"] == 1)
        ]
        singlePair["num_fragment_T1_only"] = len(fragT1Set)
        singlePair["num_fragment_T2_only"] = len(fragT2Set)
        singlePair["num_fragment_shared"] = len(fragSharedSet)
        singlePair["prop_fragment_diff"] = (
            singlePair["num_fragment_T1_only"] + singlePair["num_fragment_T2_only"]
        ) / (
            singlePair["num_fragment_T1_only"]
            + singlePair["num_fragment_T2_only"]
            + singlePair["num_fragment_shared"]
        )
        singlePair["prop_fragment_similar"] = 1 - singlePair["prop_fragment_diff"]
        singlePair["fragment_T1_only"] = "|".join(fragT1Set["fragment_id"])
        singlePair["fragment_T2_only"] = "|".join(fragT2Set["fragment_id"])
        singlePair["fragment_shared"] = "|".join(fragSharedSet["fragment_id"])
        singlePair["num_fragment_singletons_T1_only"] = len(fragT1SingSet)
        singlePair["num_fragment_singletons_T2_only"] = len(fragT2SingSet)
        singlePair["num_fragment_singletons_shared"] = len(fragSharedSingSet)

        # Count number of nt shared/different in all EF
        singlePair["num_nt_shared"] = fragSharedSet["fragment_length"].sum()
        singlePair["num_nt_T1_only"] = fragT1Set["fragment_length"].sum()
        singlePair["num_nt_T2_only"] = fragT2Set["fragment_length"].sum()
        singlePair["num_nt_diff"] = (
            singlePair["num_nt_T1_only"] + singlePair["num_nt_T2_only"]
        )
        singlePair["total_nt"] = (
            singlePair["num_nt_shared"]
            + singlePair["num_nt_T1_only"]
            + singlePair["num_nt_T2_only"]
        )
        singlePair["prop_nt_diff"] = (
            singlePair["num_nt_T1_only"] + singlePair["num_nt_T2_only"]
        ) / (singlePair["total_nt"])
        singlePair["prop_nt_similar"] = 1 - singlePair["prop_nt_diff"]

        # Count number of nt shared/different in EF only in shared ER
        singlePair["num_nt_T1_only_in_shared_ER"] = fragT1Set[
            fragT1Set["region_id"].isin(ERSharedSet)
        ]["fragment_length"].sum()
        singlePair["num_nt_T2_only_in_shared_ER"] = fragT2Set[
            fragT2Set["region_id"].isin(ERSharedSet)
        ]["fragment_length"].sum()
        singlePair["num_nt_shared_in_shared_ER"] = fragSharedSet[
            fragSharedSet["region_id"].isin(ERSharedSet)
        ]["fragment_length"].sum()
        singlePair["total_nt_in_shared_ER"] = (
            singlePair["num_nt_shared_in_shared_ER"]
            + singlePair["num_nt_T1_only_in_shared_ER"]
            + singlePair["num_nt_T2_only_in_shared_ER"]
        )
        if singlePair["total_nt_in_shared_ER"] != 0:
            singlePair["prop_nt_diff_in_shared_ER"] = (
                singlePair["num_nt_T1_only_in_shared_ER"]
                + singlePair["num_nt_T2_only_in_shared_ER"]
            ) / (singlePair["total_nt_in_shared_ER"])
            singlePair["prop_nt_similar_in_shared_ER"] = (
                1 - singlePair["prop_nt_diff_in_shared_ER"]
            )
        else:
            singlePair["prop_nt_diff_in_shared_ER"] = 0
            singlePair["prop_nt_similar_in_shared_ER"] = 0
        singlePair["num_nt_T1_only_in_unique_ER"] = fragT1Set[
            ~fragT1Set["region_id"].isin(ERSharedSet)
        ]["fragment_length"].sum()
        singlePair["num_nt_T2_only_in_unique_ER"] = fragT2Set[
            ~fragT2Set["region_id"].isin(ERSharedSet)
        ]["fragment_length"].sum()

        # Get IR distance values
        singlePair["num_IR_fragment_T1"] = len(
            fragT1Set[fragT1Set["ef_ir_flag"].map(int) == 1]
        )
        singlePair["num_IR_fragment_T2"] = len(
            fragT2Set[fragT2Set["ef_ir_flag"].map(int) == 1]
        )
        singlePair["IR_fragment_T1"] = "|".join(
            fragT1Set[fragT1Set["ef_ir_flag"].map(int) == 1]["fragment_id"]
        )
        singlePair["IR_fragment_T2"] = "|".join(
            fragT2Set[fragT2Set["ef_ir_flag"].map(int) == 1]["fragment_id"]
        )
    return singlePair


def set_AS_flags(singlePair, ef_df, tx1_name, tx2_name, has_junctions):
    # Flag alternative splicing events

    # Flag transcripts with no shared nucleotides (nonoverlapping)
    singlePair["flag_no_shared_nt"] = np.where(singlePair["prop_nt_diff"] == 1, 1, 0)

    # Flag full-splice matches (FSM, share all junctions)
    # NOTE: monoexon transcripts can be FSM
    singlePair["flag_FSM"] = np.where(singlePair["prop_junction_diff"] == 0, 1, 0)

    # Flag incomplete-splice matches
    # (ISM, one set of junctions is a complete consecutive subset of the other)
    # NOTE: ISM flags are 0 if transcripts are FSM or if at least one is monoexon
    if has_junctions:
        if (
            singlePair["flag_FSM"] == 1
            or singlePair["junction_T1_all"] == ""
            or singlePair["junction_T2_all"] == ""
        ):
            singlePair[
                ["flag_T1_ISM_of_T2", "flag_T2_ISM_of_T1"]
            ] = 0
        else:
            singlePair["flag_T1_ISM_of_T2"] = np.where(
                singlePair["junction_T1_all"] in singlePair["junction_T2_all"],
                1,
                0
            )
            singlePair["flag_T2_ISM_of_T1"] = np.where(
                singlePair["junction_T2_all"] in singlePair["junction_T1_all"],
                1,
                0
            )

    # If a pair contains at least one IR event - 5' and 3' variation calculated
    #   but flag_alt_exon and flag_alt_donor_acceptor set to 0
    # Intron retention
    singlePair["flag_IR"] = np.where(
        singlePair["num_IR_fragment_T1"] + singlePair["num_IR_fragment_T2"] > 0, 1, 0
    )

    # 5' and 3' end variation
    t1_min_start = ef_df[ef_df["transcript_id"].str.contains(tx1_name)][
        "ef_start"
    ].min()
    t2_min_start = ef_df[ef_df["transcript_id"].str.contains(tx2_name)][
        "ef_start"
    ].min()
    t1_max_end = ef_df[ef_df["transcript_id"].str.contains(tx1_name)]["ef_end"].max()
    t2_max_end = ef_df[ef_df["transcript_id"].str.contains(tx2_name)]["ef_end"].max()
    ef_strand = ef_df["ef_strand"].unique()[0]
    # 5' variation: where difference in start if + strand or difference in end if - strand
    singlePair["flag_5_variation"] = np.where(
        ((ef_strand == "+") & (t1_min_start != t2_min_start))
        | ((ef_strand == "-") & (t1_max_end != t2_max_end)),
        1,
        0,
    )
    # 3' end variation: where difference in end if + strand or difference in start if - strand
    singlePair["flag_3_variation"] = np.where(
        ((ef_strand == "-") & (t1_min_start != t2_min_start))
        | ((ef_strand == "+") & (t1_max_end != t2_max_end)),
        1,
        0,
    )

    # If a pair contains at least one IR event - 5' and 3' variation calculated
    #   but flag_alt_exon and flag_alt_donor_acceptor set to 0
    if singlePair["flag_IR"] == 1:
        singlePair["flag_alt_exon"] = 0
        singlePair["flag_alt_donor_acceptor"] = 0
    else:
        # Alternate exons are when not all exon regions are shared
        singlePair["flag_alt_exon"] = np.where(singlePair["prop_ER_diff"] > 0, 1, 0)

        # Alternate donor/acceptors in shared ER (where not all junctions are shared)
        singlePair["flag_alt_donor_acceptor"] = np.where(
            (singlePair["prop_nt_diff_in_shared_ER"] > 0)
            & (singlePair["flag_FSM"] == 0),
            1,
            0,
        )

    # If transcripts are nonoverlapping then 5'/3' variation and alt exon flags set to 0
    if singlePair["flag_no_shared_nt"] == 1:
        singlePair["flag_alt_exon"] = 0
        singlePair["flag_5_variation"] = 0
        singlePair["flag_3_variation"] = 0
        singlePair["flag_FSM"] = 0
        singlePair["flag_T1_ISM_of_T2"] = 0
        singlePair["flag_T2_ISM_of_T1"] = 0

    return singlePair



#!/usr/bin/env python
# coding: utf-8

import pandas as pd
import numpy as np
import os
import matplotlib.pyplot as plt
import matplotlib
import seaborn as sns
import math


def calculate_donor_ratio_newINF(unit, feature, count_col, data):
    df = data.groupby([unit, feature])[count_col].count().reset_index().rename(columns = {count_col : "num_col_id"})
    df["flag_donor_acceptor"] = np.where(df["num_col_id"] > 1, 1, 0)
    df2 = df.groupby(unit).agg({feature : "count", "flag_donor_acceptor" : 'sum'})
    res = df2["flag_donor_acceptor"]/df2[feature]
    return res

def uniqueList(test_list):
    res = []
    for i in test_list:
        if i not in res:
            res.append(i)
    return res

def calculate_number(unit, feature, data, calLen = True, numbers = True):
    nr = data.shape[0]
    temp = {}
    dic_len = {}
    if calLen:
        startcol = feature.replace("_id", "_start")
        stopcol = feature.replace("_id", "_end")
    for i in range(0, nr):
        uni = data[unit][i]
        fea = data[feature][i].split("|")
        if calLen:
            length = data[stopcol][i] - data[startcol][i]
        else:
            length = 0
        if uni not in temp:
            dic_len[uni] = length
            temp[uni] = fea
        else:
            dic_len[uni] += length
            for j in fea:
                if j not in temp[uni]:
                    temp[uni].append(j)

    for key in temp:
        temp[key] = uniqueList(temp[key])                

    if numbers:             
        for key in temp:
            temp[key] = len(temp[key])

    return pd.Series(temp), pd.Series(dic_len)


def calculate_IR_ratio(data_er, irs):
    dat, length = calculate_number("gene_id", "er_transcript_ids", data_er, calLen=False, numbers = False)
    dat = pd.DataFrame(dat)
    dat = dat.reset_index()
    dat.columns = ["gene_id", "er_transcript_ids"]
    dat = dat.explode("er_transcript_ids")
    
    mergedDF = pd.merge(dat, irs, how = 'outer', on = 'er_transcript_ids', indicator = 'merge_check')
    mergedDF['flag_IR_transcript'] = np.where(mergedDF['merge_check'] == 'both', 1, 0)
    mergedDF['flag_all_transcript'] = 1
    num_transcript = mergedDF.groupby("gene_id")['flag_all_transcript'].sum()
    num_IR_transcript = mergedDF.groupby("gene_id")['flag_IR_transcript'].sum()
    df_ir_ratio = pd.DataFrame({"gene_id": num_transcript.index,
                                "num_transcript": num_transcript,
                                "num_IR_transcript": num_IR_transcript,
                                "prop_IR_transcript": num_IR_transcript/num_transcript}).reset_index(drop=True)
    return df_ir_ratio


def calculate_length(dat, feature):
    start = feature + '_start'
    stop = feature + '_end'
    ids = feature + "_id"
    res = dat[stop] - dat[start]
    res.index = dat[ids]
    return res

def generate_bins(data, nb = 10, merge_tail = 99.5, ceil = True, hardcut=False):
    bins = [min(data)]
    if not hardcut:
        cut = np.percentile(data, merge_tail)
        itv = (cut - min(data))/(nb-1)
    else:
        cut = 100
        if max(data) > cut:
            itv = (cut - min(data))/(nb-1)
        else:
            cut = np.percentile(data, merge_tail)
            itv = (cut - min(data))/(nb-1)
    
    for i in range(1, nb):
        bins.append(min(data) + i*itv)
    
    if not hardcut:
        bins = bins + [max(data)+0.1]
    else:
        if max(data) > cut:
            bins = bins + [max(data)+0.1]
        else:
            bins = bins + [100]

    if ceil:
        bins = [math.ceil(x) for x in bins]
    return bins


def set_boxAttri(box, color_list):
    for patch, filers, whis, med, color in zip(box['boxes'], box['fliers'], box['whiskers'], box['medians'], color_list):
        patch.set_color(color)
        filers.set_markeredgecolor(color)
        whis.set_color(color)
        med.set_color('black')
    return

def calculate_varFeatureRatio(varList, allList):
    df1 = allList.to_frame()
    df1.columns = ['n']
    df1.insert(0, 'name', allList.index)
    df2 = varList.to_frame()
    df2.columns = ['n']
    df2.insert(0, 'name', varList.index)
    df_merge = df1.merge(df2, how='left', on = 'name', suffixes=('_totalFeature', '_varFeature')).fillna(0)
    df_merge['ratio'] = df_merge['n_varFeature'] / df_merge['n_totalFeature']
    return df_merge

def hist_plot(dat, bins, colors, text=True, omitText = []):
    x_coords = []
    y_coords = []
    first_height = len(dat[dat.values < bins[1]])
    #second_height = len(dat[(dat.values >= bins[1]) & (dat.values < bins[2])])
    #ylim = max(first_height, second_height) * 1.1
    xlim = len(bins) * 1.1
    shrink = (len(bins)-1)/10
    plt.xlim(1, xlim)
    plt.plot(xlim, first_height)
    plt.grid(False)
    plt.xticks([])
    omitIter = 0
    for i in range(0, len(bins)-1):
        x = i+2
        y = len(dat[(dat.values < bins[i+1]) & (dat.values >= bins[i])])
        col = colors[i]
        plt.vlines(x, 0, y, color = col, linewidth=44/shrink)
        if text and (omitIter not in omitText):
            plt.text(x, y, y, fontsize = 10/shrink, horizontalalignment='center',verticalalignment='bottom')
        x_coords.append(x)
        y_coords.append(y)
        omitIter+=1
    plt.ylim(0, max(y_coords)*1.1)
    return x_coords, y_coords



def counts_projectToColor(lowCol, highCol, counts, firstBinColor = False):
    sumC = sum(counts)
    if sumC == 0:
        return tuple(map(lambda x: x/255 ,list(lowCol)))
    if firstBinColor:
        return tuple(map(lambda x: x/255, list(lowCol)))
    ratio = [i/sumC for i in counts]
    heat_colors = [((lowCol[0] + (highCol[0] - lowCol[0]) * k)/255, (lowCol[1] + (highCol[1] - lowCol[1]) * k)/255, (lowCol[2] + (highCol[2] - lowCol[2]) * k)/255) for k in ratio]
    return heat_colors



def heatmap_ratio_visualize(data, ax, heat_nrows, tick_pos, lowCol=(0, 255, 0), highCol=(255, 0, 0), kb = 10):
    heat_bins = generate_bins([0,1], nb = heat_nrows, merge_tail=100, ceil=False)
    heat_x_coords = heat_bins[0:heat_nrows]
    heat_x_coords = [e+0.05 for e in heat_x_coords]
    heat_starts = tick_pos[0:(kb+1)]
    heat_ends = tick_pos[1:(kb+2)]
    res = []
    for b in range(0, kb+1):
        ind = pd.Series(data[b])
        counts = []
        for i in range(0, len(heat_bins)-1):
            each = ind[(ind.values < heat_bins[i+1]) & (ind.values >= heat_bins[i])]
            counts.append(each.count())
        if b == 0:
            firstBinColor = True
        else:
            firstBinColor = False
        heat_colors = counts_projectToColor(lowCol, highCol, counts, firstBinColor = firstBinColor)
        ax.hlines(heat_x_coords, heat_starts[b], heat_ends[b], color = heat_colors, linewidth = 65/heat_nrows)
        res.append(counts)
    return res


def find_break_lim(y_coords):
    n = len(y_coords)
    sorted_ys = np.flip(np.sort(y_coords))
    sorted_idx = np.flip(np.argsort(y_coords))
    differences = sorted_ys[0:(n-1)] - sorted_ys[1:n]
    max_idx = np.argmax(differences)
    break_y_up = y_coords[sorted_idx[max_idx]]
    break_y_down = y_coords[sorted_idx[max_idx+1]]
    return break_y_up, break_y_down

def extract_genes_from_bins(geneList, data, results, types):
    if types == 'feature':
        each = data[data.index.isin(geneList)].tolist()
        each = [x for x in each if str(x) != 'nan']
    elif types == "ratio":
        each = data[data['name'].isin(geneList)]['ratio'].tolist()
    results.append(each)
    

def visualize_combined(list_main, list_e, list_eu, df_ratio_IR, f_donor_ratio, df_ratio_e, df_ratio_f_lensum, prefix, out, kbins=4):
    dat = list_main
    dat2 = list_e
    dat3 = list_eu

    xlabel = 'Number of transcripts per gene'
    ylabel = 'Number of genes'
    ylabel2 = 'Exon regions\nper gene'
    ylabel3 = 'Unique exons\nper gene'
    ylabel4 = 'Proportion of transcripts\nwith intron retention'
    ylabel5 = 'Proportion of exon regions\nwith alt. don/acc'
    ylabel6 = 'Proportion of variable\nexon region'
    ylabel7 = 'Proportion of variable \n nucleotides'
    titile = ''
    #lowColor = (0, 255, 0)
    lowColor = (224, 224, 224)
    highColor = (255, 0, 0)

    font2 = {'family' : 'serif',
             'weight' : 'normal',
             'size'   : 13,
    }
    font3 = {'family' : 'serif',
             'weight' : 'normal',
             'size'   : 8,
    }
    if max(dat) > 16:
        lastbin = max(dat)
    else:
        lastbin = 17
    bins = [1, 2, 4, 8, 16, lastbin]

    ticklinewidth = 0.5
    colors = sns.color_palette("hls", kbins+1)
    xlim = (kbins + 2) * 1.1
    shrink = (len(bins)-1)/10
    
    sns.set_style("whitegrid")
    fig = plt.figure(figsize=(13,10), facecolor='white',edgecolor='black')
    grid = plt.GridSpec(9, 1, wspace=0, hspace=0.15)
    y_coords = []
    x_coords = []
    ###############################
    # plot section 1
    ax1 = fig.add_subplot(grid[0,0])
    ax1.patch.set_facecolor('white')
    ax1.axes.get_yaxis().set_visible(True)
    ax1.spines['bottom'].set_visible(False)
    ax1.set_title(titile, fontsize=14)

    x_coords, y_coords = hist_plot(dat, bins, colors, text=False)

    break_y_up, break_y_down = find_break_lim(y_coords)
    
    ax1.set_ylim(break_y_up * 0.8, max(y_coords) * 1.2)
    for i in range(0, len(bins)-1):
        if y_coords[i] >= break_y_up:
            ax1.text(x_coords[i], y_coords[i], y_coords[i], fontsize=10/shrink,
                     horizontalalignment='center', verticalalignment='bottom')

    tick_pos = [1] + x_coords
    tick_pos = [i+0.5 for i in tick_pos]
    ax1.tick_params(axis='y', which='major', pad=-5, labelsize=6, direction='in')
    for tick in ax1.yaxis.get_majorticklabels():
        tick.set_horizontalalignment("left")

    ##################################
    # plot section 2
    ax2 = fig.add_subplot(grid[1:3,0])
    ax2.patch.set_facecolor('white')
    ax2.spines['top'].set_visible(False)
    ax2.spines['bottom'].set_visible(False)
#    ax2.set_ylabel(ylabel,font2)
    ax2.set_ylabel(ylabel, font3, labelpad = 50, rotation=20, verticalalignment='top')
    #ax2.plot(x_coords, y_coords, color = 'steelblue', lw = 1) # fit a line
    
    hist_plot(dat, bins, colors, text=False)
    ax2.set_ylim(0, break_y_down * 1.3)
    for i in range(0, len(bins)-1):
        if y_coords[i] <= break_y_down:
            ax2.text(x_coords[i], y_coords[i], y_coords[i], fontsize = 10/shrink, horizontalalignment='center',verticalalignment='bottom')

    ax2.tick_params(axis='y', which='major', pad=-5, labelsize=6, direction='in')
    for tick in ax2.yaxis.get_majorticklabels():
        tick.set_horizontalalignment("left")
    ################################
    # extract genes based on bins
    feature2s = []
    feature3s = []
    ratio_IRs = []
    ratio_donors = []
    ratio_es = []
    ratio_flensum = []
    for i in range(0, len(bins)-1):
        genes = dat[(dat.values < bins[i+1]) & (dat.values >= bins[i])].index.tolist()        
        extract_genes_from_bins(genes, dat2, feature2s, 'feature')
        extract_genes_from_bins(genes, dat3, feature3s, 'feature')
        extract_genes_from_bins(genes, df_ratio_IR, ratio_IRs, 'ratio')
        extract_genes_from_bins(genes, f_donor_ratio, ratio_donors, 'ratio')
        extract_genes_from_bins(genes, df_ratio_e, ratio_es, 'ratio')
        extract_genes_from_bins(genes, df_ratio_f_lensum, ratio_flensum, 'ratio')
        
    ################################
    # plot section 3
    ax3 = fig.add_subplot(grid[3,0])
    ax3.patch.set_facecolor('white')
    ax3.spines['top'].set_visible(False)
    ax3.spines['bottom'].set_visible(False)
    ax3.grid(False)
    ax3.set_ylim(0, max(dat2))
    ax3.set_ylabel(ylabel2, font3, labelpad = 50, rotation=20, verticalalignment='top')
    box = ax3.boxplot(feature2s, positions = x_coords, notch = False, patch_artist=True)
    set_boxAttri(box, colors)
    
    y_arrow = max(dat2)
    ax3.set_xlim(1, xlim)
    ax3.set_xticklabels([])
    ax3.vlines(tick_pos, 0, max(dat2), color = 'grey', linewidth=ticklinewidth)
    ax3.set_yticks([y_arrow * 0.1, round(y_arrow/2), y_arrow*0.95])
    ax3.set_yticklabels([0, round(y_arrow/2), y_arrow], ha = 'left')
    ax3.tick_params(axis='y', which='major', pad=-5, labelsize=6, direction='in')

    for i in range(0, len(bins)-1): 
        ax3.arrow(x_coords[i], y_arrow, 0, -y_arrow/20, overhang=y_arrow/20, head_width=0.2, head_length=1, width = 1, shape="full",fc=colors[i], ec=colors[i],alpha=0.9)


    ##################################
    # plot section 4
    ax4 = fig.add_subplot(grid[4,0])
    ax4.patch.set_facecolor('white')
    ax4.spines['top'].set_visible(False)
    ax4.spines['bottom'].set_visible(False)
    ax4.grid(False)
    ax4.set_ylim(0, max(dat3))
    ax4.set_ylabel(ylabel3, font3, labelpad = 50, rotation=20, verticalalignment='top')
    box2 = ax4.boxplot(feature3s, positions = x_coords, notch = False, patch_artist=True)
    set_boxAttri(box2, colors)

    y_arrow = max(dat3)
    ax4.set_xticklabels([])
    ax4.set_xlim(1, xlim)
    ax4.set_yticks([y_arrow * 0.1, round(y_arrow/2), y_arrow*0.95])
    ax4.set_yticklabels([0, round(y_arrow/2), y_arrow], ha = 'left')
    ax4.tick_params(axis='y', which='major', pad=-5, labelsize=6, direction='in')
    
    ax4.vlines(tick_pos, 0, max(dat3), color = 'grey', linewidth=ticklinewidth)
    for i in range(0, len(bins)-1): 
        ax4.arrow(x_coords[i], y_arrow, 0, -y_arrow/20, overhang=y_arrow/20, head_width=0.2, head_length=1, width = 1, shape="full",fc=colors[i], ec=colors[i], alpha=0.9)


    
    ###########################################
    # set heatmap resolution
    heat_nrows = 20
    heat_ylim = 1.1
    
    ####### plot section 5
    
    for hm, ratio, lab in zip([5, 6, 7], [ratio_IRs, ratio_donors, ratio_es], [ylabel4, ylabel5, ylabel6]):
        
        ax = fig.add_subplot(grid[hm,0])
        ax.patch.set_facecolor('white')
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.grid(False)
        ax.set_ylim(0, heat_ylim)
        
        heatmap_ratio_visualize(ratio, ax, heat_nrows, tick_pos, kb = kbins, lowCol = lowColor, highCol = highColor)
        ax.set_ylabel(lab, font3, labelpad = 60, rotation=20, verticalalignment='top')
        ax.set_yticks([0.1, 0.5, 0.95])
        ax.set_yticklabels([0, 0.5, 1], ha = 'left')
        ax.tick_params(axis='y', which='major', pad=-5, labelsize=6, direction='in')
        
        ax.set_xticklabels([])
        ax.set_xlim(1, xlim)
        ax.vlines(tick_pos, 0, heat_ylim-0.03, color = '#4682b4', linewidth=ticklinewidth)
        ax.hlines(0, tick_pos[0], tick_pos[-1], color = '#D3D3D3', linewidth=ticklinewidth*3)

    
     ##################################
    ####### plot section 6
    ax6 = fig.add_subplot(grid[8,0])
    ax6.patch.set_facecolor('white')
    ax6.spines['top'].set_visible(False)
    ax6.grid(False)
    ax6.set_ylim(0, heat_ylim)
    
    heatmap_ratio_visualize(ratio_flensum, ax6, heat_nrows, tick_pos, kb = kbins, lowCol = lowColor, highCol = highColor)
    
    ax6.set_xlim(1, xlim)
    ax6.set_xlabel(xlabel, font2, labelpad = 10) 
    ax6.set_xticks(tick_pos)
    ax6.set_xticklabels([str(round(i, 2)) for i in bins])
    
    ax6.set_ylabel(ylabel7, font3, labelpad = 50, rotation=20, verticalalignment='top')
    ax6.set_yticks([0.1, 0.5, 0.95])
    ax6.set_yticklabels([0, 0.5, 1], ha = 'left')
    ax6.tick_params(axis='y', which='major', pad=-5, labelsize=6, direction='in', length=1)
    ax6.vlines(tick_pos, 0, heat_ylim-0.03, color = '#4682b4', linewidth=ticklinewidth)
    
#    outfile = out + "/transcriptome_summary_plot.pdf"
    if prefix is not None:
        outfile = "{}/{}_transcriptome_summary_plot.png".format(out, prefix)
    else:
        outfile = "{}/transcriptome_summary_plot.png".format(out)
#    plt.savefig(outfile, dpi=600, format='pdf')
    plt.savefig(outfile, dpi=600, format='png')
    return

def plot_transcriptome(er_data, ef_data, ir_data, uniqex_data, outdir, prefix):
    er_data['er_start'] = er_data['er_start'].astype(int)
    er_data['er_end'] = er_data['er_end'].astype(int)
    ef_data['ef_start'] = ef_data['ef_start'].astype(int)
    ef_data['ef_end'] = ef_data['ef_end'].astype(int)
    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.rcParams['ps.fonttype'] = 42
    ###############
    ####### process gene transcript input using er input

    num_list_main, g_lensum = calculate_number('gene_id', 'er_transcript_ids', er_data, calLen = False)

    ###############
    ###### process exon region input
    er_data_var = er_data[er_data['er_annotation_frequency'] != 'constitutive'].reset_index().drop('index', axis=1)

    num_list_e, e_lensum = calculate_number('gene_id', 'er_id', er_data)
    num_list_e_varXcript, e_lensum_varXcript = calculate_number('gene_id', 'er_id', er_data_var)

    df_ratio_e = calculate_varFeatureRatio(num_list_e_varXcript, num_list_e)
    ################
    ###### process fragment input
    ef_data_var = ef_data[ef_data['ea_annotation_frequency'] != 'constitutive'].reset_index().drop('index', axis=1)


    # calculate donor/acceptor ratio for each gene
    f_donor_ratio = pd.DataFrame()
    seri_fs = calculate_donor_ratio_newINF('gene_id', 'er_id', 'ef_id', ef_data)
    f_donor_ratio['name'] = seri_fs.index
    f_donor_ratio['ratio'] = seri_fs.values

    num_list_f, f_lensum = calculate_number('gene_id', 'ef_id', ef_data)
    num_list_f_varXcript, f_lensum_varXcript = calculate_number('gene_id', 'ef_id', ef_data_var)
    del(ef_data)
    df_ratio_f_lensum = calculate_varFeatureRatio(f_lensum_varXcript, f_lensum)
    ##################

    num_list_ue = uniqex_data.set_index('gene_id')['num_uniq_exon']

    df_ratio_IR = calculate_IR_ratio(er_data, ir_data)
    df_ratio_IR.columns = ['name', 'num_transcript', 'num_IR_transcript', 'ratio']
    del(er_data)

    visualize_combined(
        num_list_main,
        num_list_e,
        num_list_ue,
        df_ratio_IR,
        f_donor_ratio,
        df_ratio_e,
        df_ratio_f_lensum,
        prefix,
        out = os.path.normpath(outdir),
    )









#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 26 11:33:04 2021

@author: adalena.nanni
"""

import matplotlib.pyplot as plt

# Import plot functions
from . import plot_functions as PF


def plot_two_gtf_pairwise(outdir, md_data, f1_odds, f2_odds, name1, name2, prefix):
    """
    Plot 2 GTF Pairwise output including the following:

        1) Pie charts for the number of genes where the transcripts are the same,
            where one dataset has more transcripts that the onther, and where
            genes are exclusive (pie chart for all genes, single transcript in
            at least one dataset, multi-transcript in at least one dataset, and
            multi-transcript in both datasets)

        2) UpSet plot for number of transcript in genes

        3) A stacked bar chart of the genes with reciprocal minimum pairs for
            all transcript pairs possible (Match:Reciprocal Pairs or Greater:Reciprocal Pairs),
            only some of the transcript pairs (Match:Partial Reciprocal Pairs or Greater:Partial Reciprocal Pairs),
            or none of the transcript pars (Match:No Reciprocal Pairs or Greatre:No Reciprocal Pairs)

        4) The proportional stacked bar chart of (3)

        5) For all minimum pairs of transcripts in either dataset: Upset plot
            for the number of transcripts with each kind of alternative splicing (AS)
            and the distribution of nt differences between the pairs plotted

        6) For minimum pair transcripts of first dataset: Upset plot
            for the number of transcripts with each kind of alternative splicing (AS)
            and the distribution of nt differences between the pairs plotted

        7) For minimum pair transcripts of second dataset: Upset plot
            for the number of transcripts with each kind of alternative splicing (AS)
            and the distribution of nt differences between the pairs plotted
             
        8) For all reciprocal minimum pair transcripts: Upset plot for the number
            of transcripts with each kind of alternative splicing (AS)
            and the distribution of nt differences between the pairs plotted

        9) For all extra minimum pair transcripts: Upset plot for the number
            of transcripts with each kind of alternative splicing (AS)
            and the distribution of nt differences between the pairs plotted
            
        10) Upset plot for number of genes with each kind of AS when comparing the
            reciprocally min match pairs of the two datasets with added box plots
            of the number of min match pairs present in the gene, the number of
            transcripts in each dataset, the average number of nt different,
            and the average proportion of nt different

        11) Plot average number of nt different in recip. min match pairs against
            avg number of nt different in min pairs that are extra
            (either in the greater sets of in set that do not match)

        12) Same as plot (11) with zoom in axises to the mean value of non-zero
            nucleotide differences (max of the two means from recip. min pairs and extras)
    """
# Plot 1
    # Plot pie chart for number of genes where:
    #   the numbers of transcripts are the same (Match) between the two datasets,
    #   where one dataset has more transcripts that the other,
    #   or where one dataset has transcripts and the other has none for the gene

    # Plot pie charts split by 1) all genes, 2) single transcript in at least 1 dataset,
    #   3) multi-transcript in at least 1 dataset, and 4) multi-transcript in both datasets
    PF.plot_transcript_in_gene_split_pie(md_data, f1_odds, f2_odds, name1,
                                         name2, "{}/{}_transcript_in_gene_split_pie.rtf".format(outdir, prefix))
    plt.savefig("{}/{}_transcript_in_gene_split_pie.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 2
    # Upset plot for number of transcripts in genes
    PF.plot_transcript_in_gene_upset(md_data, f1_odds, f2_odds, name1, name2,
                                     "{}/{}_transcript_in_gene_upset.rtf".format(outdir, prefix))
    plt.savefig("{}/{}_transcript_in_gene_upset.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 3
    # Plot stacked bar chart of genes with reciprocal minimum pairs for:
    #   1) all transcript pairs possible (Match:Reciprocal Pairs or Greater:Reciprocal Pairs)
    #   2) only some of the transcript pairs (Match:Partial Reciprocal Pairs or Greater:Partial Reciprocal Pairs)
    #   3) none of the transcript pars (Match:No Reciprocal Pairs or Greater:No Reciprocal Pairs)
    PF.plot_gene_stack(md_data, name1, name2, "{}/{}_transcript_in_gene_stackCount.rtf".format(outdir, prefix))
    plt.savefig("{}/{}_transcript_in_gene_stackCount.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 4
    # With percentage of genes instead of count
    PF.plot_gene_stack(md_data, name1, name2,
                       "{}/{}_transcript_in_gene_stack_proportion.rtf".format(outdir, prefix),
                       useProp=True)
    plt.savefig("{}/{}_transcript_in_gene_stack_proportion.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 5
    # For all minimum pairs of transcripts in either dataset:
    # Upset plot for the number of transcripts with each kind of alternative splicing (AS)
    #   and the distribution of nt differences between the pairs plotted
    PF.plot_min_pair_AS_upset_nt_box(md_data, name1, name2,
                                           "{}/{}_all_min_pair_AS_upset_nt_box.rtf".format(outdir, prefix),
                                           reciprocal=False, pairs="all")
    plt.savefig("{}/{}_all_min_pair_AS_upset_nt_box.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 6
    # For minimum pair transcripts of first dataset:
    # Upset plot for the number of transcripts with each kind of alternative splicing (AS)
    #   and the distribution of nt differences between the pairs plotted
    PF.plot_min_pair_AS_upset_nt_box(md_data, name1, name2,
                                           "{}/{}_min_pair_AS_upset_nt_box.rtf".format(outdir, name1),
                                           reciprocal=False, pairs="first")
    plt.savefig("{}/{}_min_pair_AS_upset_nt_box.png".format(outdir, name1), dpi=600, format="png")
    plt.clf()

# Plot 7
    # For minimum pair transcripts of second dataset:
    # Upset plot for the number of transcripts with each kind of alternative splicing (AS)
    #   and the distribution of nt differences between the pairs plotted
    PF.plot_min_pair_AS_upset_nt_box(md_data, name1, name2,
                                           "{}/{}_min_pair_AS_upset_nt_box.rtf".format(outdir, name2),
                                           reciprocal=False, pairs="second")
    plt.savefig("{}/{}_min_pair_AS_upset_nt_box.png".format(outdir, name2), dpi=600, format="png")
    plt.clf()

# Plot 8
    # For all reciprocal minimum pair transcripts:
    # Upset plot for the number of transcripts with each kind of AS and the distribution
    #   of nt differences between the pairs plotted above
    PF.plot_min_pair_AS_upset_nt_box(md_data, name1, name2,
                                           "{}/{}_recip_min_pair_AS_upset_nt_box.rtf".format(outdir, prefix),
                                           reciprocal=True)
    plt.savefig("{}/{}_recip_min_pair_AS_upset_nt_box.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 9
    # For all extra minimum pair transcripts:
    # Upset plot for the number of transcripts with each kind of AS and the distribution
    #   of nt differences between the pairs plotted above
    PF.plot_min_pair_AS_upset_nt_box(md_data, name1, name2,
                                           "{}/{}_extra_min_pair_AS_upset_nt_box.rtf".format(outdir, prefix),
                                           reciprocal=False)
    plt.savefig("{}/{}_extra_min_pair_AS_upset_nt_box.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 10
    # Upset plot for number of genes with each kind of AS when comparing the
    #   reciprocally min match pairs of the two datasets with added box plots
    #   of the number of min match pairs present in the gene,
    #   the number of transcripts in each dataset, the avg number of nt different,
    #   and the avg proportion of nt different
    PF.plot_gene_recip_min_AS_upset(md_data, name1, name2,
                                           "{}/{}_recip_min_gene_AS_upset.rtf".format(outdir, prefix))
    plt.savefig("{}/{}_recip_min_gene_AS_upset.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 11
    # Plot average number of nt different in recip. min match pairs against
    #   avg number of nt different in min pairs that are extra
    #   (either in the greater sets of in set that do not match)
    PF.plot_gene_avg_nt_diff_pairs(md_data, name1, name2, "{}/{}_gene_avg_nt_diff_pairs.rtf".format(outdir, prefix))
    plt.savefig("{}/{}_gene_avg_nt_diff_pairs.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()

# Plot 12
    # Zoom in axises to the mean value of non-zero nt differences (max of the two means from recip. min pairs and extras)
    PF.plot_gene_avg_nt_diff_pairs(md_data, name1, name2,
                                   "{}/{}_gene_avg_nt_diff_pairs_zoomIn.rtf".format(outdir, prefix),
                                   zoomMean=True)
    plt.savefig("{}/{}_gene_avg_nt_diff_pairs_zoomIn.png".format(outdir, prefix), dpi=600, format="png")
    plt.clf()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jun 12 16:13:11 2021

@author: adalena.nanni
"""

import matplotlib.pyplot as plt
from loguru import logger

# Import plot functions
from . import plot_functions as PF

# Import transcriptome summary plot functions
from . import plot_transcriptome as PT


def plot_one_gtf_gene(er_data, ef_data, ir_data, uniqex_data, outdir, prefix):
    """
    Plot 1 GTF Pairwise output including the following:

        1) Upset plot for the number of genes with each kind of
            alternative splicing (AS) and the distribution of the average
            nt differences between transcript pairs within each gene
    """
# Plot 1
    # Summary of structural elements of transcriptome, grouped by number of
    #   transcripts per gene
    PT.plot_transcriptome(er_data, ef_data, ir_data, uniqex_data, outdir, prefix)
    plt.clf()
    # Make legend for plot 1
    # !!! Move this to plot functions when plot_transcriptome is moved
    if prefix is not None:
        legendOut = "{}/{}_transcriptome_summary_plot.rtf".format(outdir, prefix)
    else:
        legendOut = "{}/transcriptome_summary_plot.rtf".format(outdir)
    legendText = (
        "Summary of structural elements from input GTF. Structural elements "
        "include the frequency of genes within each range of transcript per "
        "gene values (top row, histogram). For each range of transcripts per "
        "gene, the distribution of exon regions per gene (second row, "
        "box plot) represent the number of exons disregarding alternative "
        "donors and acceptors, the number of unique exons per gene (third row, "
        "box plot) represents alternative donor/acceptors and alternative "
        "exons, the proportion of transcripts with intron retention "
        "(fourth row, heatmap), the proportion of exon regions with alternate "
        "donors/acceptors (fifth row, heatmap), the proportion of variable "
        "exon regions (sixth row, heatmap) represents the alternative exons "
        "disregarding alternative donors and acceptors, and the proportion of "
        "variable nucleotides (bottom row, heatmap) represents nucleotides "
        "that are contained in alternative exons and alternative donors and "
        "acceptors."
    )
    with open(legendOut, "w") as outFile:
        PF.start_rtf(outFile)
        outFile.write(
            r"\b Figure. Transcriptome summary \b0 \line {}".format(
                legendText
            )
        )
        PF.end_rtf(outFile)

# Plot 2
    # Distribution of proportion of nucleotide variability across all genes
    try:
        if prefix is not None:
            PF.plot_gene_prop_nt_variablility(ef_data,"{}/{}_all_gene_prop_nt_variablility.rtf".format(outdir, prefix))
            plt.savefig("{}/{}_all_gene_prop_nt_variablility.png".format(outdir, prefix), dpi=600, format="png")
        else:
            PF.plot_gene_prop_nt_variablility(ef_data,"{}/all_gene_prop_nt_variablility.rtf".format(outdir))
            plt.savefig("{}/all_gene_prop_nt_variablility.png".format(outdir), dpi=600, format="png")

        plt.clf()
    except ValueError as e:
        logger.error("Not enough data to plot nt variability {}".format(e))
        return

# Plot 3
    # Distribution of proportion of nucleotide variability across multi-transcript
    #   genes
    if prefix is not None:
        PF.plot_gene_prop_nt_variablility(ef_data,
                                          "{}/{}_multi_xcrpt_gene_prop_nt_variablility.rtf".format(outdir, prefix),
                                          multitranscript=True)
        plt.savefig("{}/{}_multi_xcrpt_gene_prop_nt_variablility.png".format(outdir, prefix), dpi=600, format="png")
    else:
        PF.plot_gene_prop_nt_variablility(ef_data,
                                          "{}/multi_xcrpt_gene_prop_nt_variablility.rtf".format(outdir),
                                          multitranscript=True)
        plt.savefig("{}/multi_xcrpt_gene_prop_nt_variablility.png".format(outdir), dpi=600, format="png")
    plt.clf()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 11 16:46:40 2021

@author: adalena.nanni
"""

import pandas as pd
import matplotlib.pyplot as plt

# Import plot functions
from . import plot_functions as PF


def calculate_complexity(outdir, gtf_df, skip_plots, prefix=None):
    """
    Calculate complexity measures from gtf dataframe:
        1) Transcripts per gene
        2) Unique exons per gene
        3) Exons per transcript
    """
    # Get transcript and gene level dataframes using dataframe from gtf
    gtf_df["num_exon"] = 1
    gtf_df["exon_id"] = (
        gtf_df["seqname"].map(str)
        + ":"
        + gtf_df["start"].map(str)
        + ":"
        + gtf_df["end"].map(str)
        + ":"
        + gtf_df["strand"].map(str)
    )
    transcriptDF = (
        gtf_df.groupby("transcript_id").agg({"num_exon": "sum"}).reset_index()
    )
    geneDF = (
        gtf_df.groupby("gene_id")
        .agg({"transcript_id": "nunique", "num_exon": "sum", "exon_id": "nunique"})
        .reset_index()
        .rename(columns={"transcript_id": "num_transcript", "exon_id": "num_uniq_exon"})
    )

    counts = pd.DataFrame()
    # Get number of transcripts
    counts.loc[0, "num_transcript"] = geneDF["num_transcript"].sum()

    # Get number of genes
    counts.loc[0, "num_gene"] = len(geneDF)

    # Get number of exons (same exon in multiple transcripts counted multiple times)
    counts.loc[0, "num_exon"] = geneDF["num_exon"].sum()

    # Get number of unique exons (can still overlap but not have the same coordinates)
    # Must use GTF and not gene level counts since unique exons can be in multiple genes
    counts.loc[0, "num_uniqExon"] = len(
        gtf_df[["seqname", "start", "end", "strand"]].drop_duplicates()
    )

    # Get minimum values
    counts.loc[0, "min_transcriptPerGene"] = geneDF["num_transcript"].min()
    counts.loc[0, "min_exonPerTranscript"] = transcriptDF["num_exon"].min()
    counts.loc[0, "min_exonPerGene"] = geneDF["num_uniq_exon"].min()

    # Get Q1 values
    counts.loc[0, "q1_transcriptPerGene"] = geneDF["num_transcript"].quantile(0.25)
    counts.loc[0, "q1_exonPerTranscript"] = transcriptDF["num_exon"].quantile(0.25)
    counts.loc[0, "q1_exonPerGene"] = geneDF["num_uniq_exon"].quantile(0.25)

    # Get median values
    counts.loc[0, "median_transcriptPerGene"] = geneDF["num_transcript"].median()
    counts.loc[0, "median_exonPerTranscript"] = transcriptDF["num_exon"].median()
    counts.loc[0, "median_exonPerGene"] = geneDF["num_uniq_exon"].median()

    # Get Q3 values
    counts.loc[0, "q3_transcriptPerGene"] = geneDF["num_transcript"].quantile(0.75)
    counts.loc[0, "q3_exonPerTranscript"] = transcriptDF["num_exon"].quantile(0.75)
    counts.loc[0, "q3_exonPerGene"] = geneDF["num_uniq_exon"].quantile(0.75)

    # Get max values
    counts.loc[0, "max_transcriptPerGene"] = geneDF["num_transcript"].max()
    counts.loc[0, "max_exonPerTranscript"] = transcriptDF["num_exon"].max()
    counts.loc[0, "max_exonPerGene"] = geneDF["num_uniq_exon"].max()

    # Get mean values
    counts.loc[0, "mean_transcriptPerGene"] = geneDF["num_transcript"].mean()
    counts.loc[0, "mean_exonPerTranscript"] = transcriptDF["num_exon"].mean()
    counts.loc[0, "mean_exonPerGene"] = geneDF["num_uniq_exon"].mean()

    # Get std
    counts.loc[0, "std_transcriptPerGene"] = geneDF["num_transcript"].std()
    counts.loc[0, "std_exonPerTranscript"] = transcriptDF["num_exon"].std()
    counts.loc[0, "std_exonPerGene"] = geneDF["num_uniq_exon"].std()

    # Plot boxplots of complexity measures and output counts to file
    if prefix is None:
        if not skip_plots:
            PF.plot_complexity_box(
                geneDF, transcriptDF, outdir, "{}/complexity_plots.rtf".format(outdir)
            )
            plt.savefig("{}/complexity_plots.png".format(outdir), dpi=600, format="png")
            plt.clf()
        counts.to_csv(
            "{}/transcriptome_complexity_counts.csv".format(outdir), index=False
        )
        return geneDF[["gene_id", "num_uniq_exon"]]
    #        plot_complexity_violin(geneDF,transcriptDF,outdir,legendOut)
    #        plt.savefig("{}/complexity_plots_violin.png".format(outdir),dpi=600,format="png")
    else:
        if not skip_plots:
            PF.plot_complexity_box(
                geneDF,
                transcriptDF,
                outdir,
                "{}/{}_complexity_plots.rtf".format(outdir, prefix),
            )
            plt.savefig(
                "{}/{}_complexity_plots.png".format(outdir, prefix),
                dpi=600,
                format="png",
            )
            plt.clf()
        counts.to_csv(
            "{}/{}_transcriptome_complexity_counts.csv".format(outdir, prefix),
            index=False,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 26 11:33:04 2021

@author: adalena.nanni
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import seaborn as sns
from upsetplot import UpSet

#  Plot Functions


def plot_transcript_in_gene_split_pie(
    md_data, f1_odds, f2_odds, name1, name2, legendOut
):
    """
    Plot pie charts for the number of genes where the transcripts are the same,
            where one dataset has more transcripts that the onther, and where
            genes are exclusive (pie chart for all genes, single transcript in
            at least one dataset, multi-transcript in at least one dataset, and
            multi-transcript in both datasets
    """
    # Get counts
    cols = {"transcript_id": "num_transcript_in_gene_" + name1}
    geneF1only = (
        f1_odds.groupby("gene_id")["transcript_id"]
        .nunique()
        .reset_index()
        .rename(columns=cols)
    )

    geneF1only["num_transcript_in_gene_" + name2] = 0
    geneF1only["transcript_in_gene"] = name1 + "_only"
    cols = {"transcript_id": "num_transcript_in_gene_" + name2}
    geneF2only = (
        f2_odds.groupby("gene_id")["transcript_id"]
        .nunique()
        .reset_index()
        .rename(columns=cols)
    )
    geneF2only["num_transcript_in_gene_" + name1] = 0
    geneF2only["transcript_in_gene"] = name2 + "_only"

    # Make dictionary of names for plot and order
    geneCountNameDict = {
        "match": "Match",
        name1 + "_greater": name1 + " Greater",
        name2 + "_greater": name2 + " Greater",
        name1 + "_only": name1 + " Only",
        name2 + "_only": name2 + " Only",
    }

    geneCountOrderDict = {
        "match": 0,
        name1 + "_greater": 1,
        name2 + "_greater": 2,
        name1 + "_only": 3,
        name2 + "_only": 4,
    }

    # Plot pie chart
    # 1) all genes
    # 2) genes with 1 xcrpt for at least 1 dataset
    # 3) genes with >1 xcrpt for at least one dataset
    # 4) genes with > 1 xcrpt for both
    fig = plt.figure()
    for num in range(1, 5):
        ax = plt.subplot2grid((2, 2), (abs(num % 2 - 1), int(num / 3)), fig=fig)
        if num == 1:
            geneCount = pd.concat(
                [
                    md_data[["gene_id", "transcript_in_gene"]].drop_duplicates(),
                    geneF1only,
                    geneF2only,
                ],
                ignore_index=True,
                sort=False,
            )["transcript_in_gene"].value_counts(sort=False)
            title = "All Genes (n = {})".format(geneCount.sum())
            totalGenes = geneCount.sum()
            legendLabels = geneCount.reindex(list(geneCountOrderDict)).index.map(
                geneCountNameDict
            )
        elif num == 2:
            geneCount = pd.concat(
                [
                    md_data[
                        (md_data["num_transcript_in_gene_" + name1] == 1)
                        | (md_data["num_transcript_in_gene_" + name2] == 1)
                    ][["gene_id", "transcript_in_gene"]].drop_duplicates(),
                    geneF1only[geneF1only["num_transcript_in_gene_" + name1] == 1],
                    geneF2only[geneF2only["num_transcript_in_gene_" + name2] == 1],
                ],
                ignore_index=True,
                sort=False,
            )["transcript_in_gene"].value_counts(sort=False)
            title = "Genes with 1 Transcript\n(in at least one dataset, n = {})".format(
                geneCount.sum()
            )
        elif num == 3:
            geneCount = pd.concat(
                [
                    geneF1only[geneF1only["num_transcript_in_gene_" + name1] > 1],
                    geneF2only[geneF2only["num_transcript_in_gene_" + name2] > 1],
                ],
                ignore_index=True,
                sort=False,
            )["transcript_in_gene"].value_counts(sort=False)
            title = "Genes with >1 Transcript\n(exclusive to one dataset, n = {})".format(
                geneCount.sum()
            )
        else:
            geneCount = (
                md_data[
                    (md_data["num_transcript_in_gene_" + name1] > 1)
                    & (md_data["num_transcript_in_gene_" + name2] > 1)
                ][["gene_id", "transcript_in_gene"]]
                .drop_duplicates()["transcript_in_gene"]
                .value_counts(sort=False)
            )
            title = "Genes with >1 Transcript\n(in both datasets, n = {})".format(
                geneCount.sum()
            )
        geneCount.reindex(list(geneCountOrderDict)).plot(
            kind="pie",
            y="transcript_in_gene",
            labels=None,
            figsize=(12, 6),
            autopct=(lambda pct: get_pie_label(pct, geneCount)),
            colormap=ListedColormap(sns.color_palette("colorblind", 15).as_hex()),
            ax=ax,
        )
        ax.set_ylabel("")
        ax.set_title(title)
    plt.legend(bbox_to_anchor=(1.2, 1.2), loc="upper left", labels=legendLabels)
    plt.tight_layout()
    legendText = (
            "Percent of genes with equal number of transcripts between "
            "{} and {} (Match, blue), more transcripts in {} compared to {} "
            "({} Greater, dark orange), more transcripts in {} compared to {} "
            "({} Greater, gray), and genes exclusive to {} ({} Only, light "
            "orange) or {} ({} Only, pink). {} genes total.".format(
                name1,
                name2,
                name1,
                name2,
                name1,
                name2,
                name1,
                name2,
                name1,
                name1,
                name2,
                name2,
                int(totalGenes),
        )
    )
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Shared and unique transcripts per gene \b0 \line "
            r"{}".format(
                legendText
            )
        )
        end_rtf(outFile)


def get_pie_label(pct, allvals):
    """
    Set pie label with percent and counts in ()
    """
    absolute = int(round(pct / 100.0 * np.sum(allvals)))
    return "{:.1f}%\n({:d})".format(pct, absolute)


def plot_gene_stack(md_data, name1, name2, legendOut, useProp=False):
    """
    Plot a stacked bar chart of the genes with reciprocal minimum pairs for all transcript pairs
    possible (Match: Reciprocal Pairs or Greater: Reciprocal Pairs), only some of the transcript
    pairs (Match: Partial Reciprocal Pairs or Greater: Partial Reciprocal Pairs), or none of the
    transcript pars (Match: No Reciprocal Pairs or Greater: No Reciprocal Pairs) using gene counts
    or proportions of Match or Greater genes
    """
    genePairDF = (
        md_data.groupby(["transcript_in_gene", "recip_min_pair_in_gene"])["gene_id"]
        .nunique()
        .reset_index()
        .pivot_table(
            index=["transcript_in_gene"],
            columns="recip_min_pair_in_gene",
            values="gene_id",
        )
    )
    for col in ["reciprocal_pairs", "partial_reciprocal_pairs", "no_reciprocal_pairs"]:
        if col not in genePairDF.columns:
            genePairDF[col] = 0
    genePairDF = genePairDF.fillna(0)
    genePairDF = genePairDF.rename(
        columns={
            "reciprocal_pairs": "Reciprocal Pairs",
            "partial_reciprocal_pairs": "Partial Reciprocal Pairs",
            "no_reciprocal_pairs": "No Reciprocal Pairs",
        },
        index={
            name1 + "_greater": name1 + " Greater",
            name2 + "_greater": name2 + " Greater",
            "match": "Match",
        },
    )
    genePairPallete = ["#FAC748", "#1D2F6F", "#8390FA"]
    if useProp:
        legendText = (
            "The stacked bar chart is scaled to represent proportions of "
            "genes within each category. Columns are: genes with equal numbers "
            "of transcripts in {} and {} (Match; n = {}), genes with {} "
            "containing more transcripts than {} ({} Greater; n = {}), and "
            "genes with {} containing more transcripts than {} ({} Greater, "
            "n = {}). Genes with Reciprocal Pairs for all transcripts (Match) "
            "or as a complete subset are colored blue. Genes with at least one "
            "but not all transcript pairs are reciprocal minimum matches "
            "(Partial Reciprocal Pairs) are colored yellow. Genes with no "
            "pairs that are reciprocal minimum matches (No Reciprocal Pairs) "
            "are colored purple.".format(
                md_data["gene_id"].nunique(),
                name1,
                name2,
                genePairDF[genePairDF.index == "Match"].sum().sum(),
                name1,
                name2,
                name1,
                genePairDF[genePairDF.index == name1 + " Greater"].sum().sum(),
                name2,
                name1,
                name2,
                genePairDF[genePairDF.index == name2 + " Greater"].sum().sum(),
            )
        )
        genePairDF = genePairDF.div(genePairDF.sum(axis=1), axis=0)
        title = "Proportion of Genes"
    else:
        legendText = (
            "The stacked bar chart is the count of the number of genes in "
            "each category (N = {}). Columns are: genes with equal numbers of "
            "transcripts in {} and {} (Match; n = {}), genes with {} "
            "containing more transcripts than {} ({} Greater; n = {}), and "
            "genes with {} containing more transcripts than {} ({} Greater, "
            "n = {}). Genes with Reciprocal Pairs for all transcripts (Match) "
            "or as a complete subset are colored blue. Genes with at least one "
            "but not all transcript pairs are reciprocal minimum matches "
            "(Partial Reciprocal Pairs) are colored yellow. Genew with no "
            "pairs that are reciprocal minimum matches (No Reciprocal Pairs) "
            "are colored purple.".format(
                md_data["gene_id"].nunique(),
                name1,
                name2,
                genePairDF[genePairDF.index == "Match"].sum().sum(),
                name1,
                name2,
                name1,
                genePairDF[genePairDF.index == name1 + " Greater"].sum().sum(),
                name2,
                name1,
                name2,
                genePairDF[genePairDF.index == name2 + " Greater"].sum().sum(),
            )
        )
        title = "Number of Genes"
    genePairDF.plot(
        kind="bar",
        figsize=(8, 6),
        stacked=True,
        rot=45,
        colormap=ListedColormap(genePairPallete),
    )
    plt.ylabel(title)
    plt.xlabel("")
    plt.legend(bbox_to_anchor=(1.04, 1), loc="upper left")
    plt.tight_layout()
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Classification of transcript pairs \b0 \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def plot_transcript_in_gene_upset(md_data, f1_odds, f2_odds, name1, name2, legendOut):
    """
    UpSet plot for number of transcripts in genes
    """
    geneF1only = (
        f1_odds.groupby("gene_id")["transcript_id"]
        .nunique()
        .reset_index()
        .rename(columns={"transcript_id": "num_transcript_in_gene_" + name1})
    )
    geneF1only["num_transcript_in_gene_" + name2] = 0
    geneF1only["transcript_in_gene"] = name1 + "_only"
    geneF2only = (
        f2_odds.groupby("gene_id")["transcript_id"]
        .nunique()
        .reset_index()
        .rename(columns={"transcript_id": "num_transcript_in_gene_" + name2})
    )
    geneF2only["num_transcript_in_gene_" + name1] = 0
    geneF2only["transcript_in_gene"] = name2 + "_only"
    geneAll = pd.concat(
        [
            md_data[
                [
                    "gene_id",
                    "num_transcript_in_gene_" + name1,
                    "num_transcript_in_gene_" + name2,
                    "transcript_in_gene",
                ]
            ].drop_duplicates(),
            geneF1only,
            geneF2only,
        ],
        ignore_index=True,
        sort=False,
    )
    colList = []
    # Flag when genes have 1-4 transcripts
    for num in range(1, 5):
        geneAll[str(num) + " Transcript(s) " + name1] = np.where(
            geneAll["num_transcript_in_gene_" + name1] == num, True, False
        )
        colList.append(str(num) + " Transcript(s) " + name1)
        geneAll[str(num) + " Transcript(s) " + name2] = np.where(
            geneAll["num_transcript_in_gene_" + name2] == num, True, False
        )
        colList.append(str(num) + " Transcript(s) " + name2)
    # Flag when genes have 5+ transcripts
    geneAll["5+ Transcript(s) " + name1] = np.where(
        geneAll["num_transcript_in_gene_" + name1] >= 5, True, False
    )
    colList.append("5+ Transcript(s) " + name1)
    geneAll["5+ Transcript(s) " + name2] = np.where(
        geneAll["num_transcript_in_gene_" + name2] >= 5, True, False
    )
    colList.append("5+ Transcript(s) " + name2)
    legendText = (
        "Rows indicate the number of transcripts in {} and {}. Columns are "
        "separated by unique combinations of the number of transcripts in "
        "{} and {}. When a gene is not present in one of the GTF files, only "
        "a single dot will be indicated. The number of genes (N = {}) for "
        "each nonoverlapping category are displayed in a histogram.".format(
            name1,
            name2,
            name1,
            name2,
            sum(
                [
                    get_value_count(geneAll, "transcript_in_gene", name1 + "_only"),
                    get_value_count(geneAll, "transcript_in_gene", name2 + "_only"),
                    get_value_count(geneAll, "transcript_in_gene", "match"),
                    get_value_count(geneAll, "transcript_in_gene", name1 + "_greater"),
                    get_value_count(geneAll, "transcript_in_gene", name2 + "_greater"),                    
                ]
            ),
        )
    )
    plot_upset(
        geneAll.set_index(colList), "Number of Genes with Each Number of Transcripts"
    )
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Number of transcripts per gene comparison UpSet plot"
            r"\b0 \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def plot_min_pair_AS_upset_nt_box(md_data, name1, name2, legendOut, reciprocal=True, pairs=None):
    """
    For all reciprocal minimum pair transcripts or minimum pair of extra transcripts:
            Upset plot for the number of transcript pairs with each kind of
            alternative splicing (AS) and the distribution of nt differences
            between the pairs plotted
    """
    if reciprocal:
        # Plot only reciprocal minimum pairs
        minPairAS = md_data[md_data["flag_recip_min_match"] == 1].copy()
    else:
        if pairs == None:
            # Plot only minimum pairs of extras, or those without a recip min pair
            minPairAS = md_data[
                (md_data["flag_recip_min_match"] != 1)
                & (
                    (md_data["flag_min_match_" + name1] == 1)
                    | (md_data["flag_min_match_" + name2] == 1)
                )
            ].copy()
        elif pairs == "all":
            minPairAS = md_data[
                    (md_data["flag_min_match_" + name1] == 1)
                    | (md_data["flag_min_match_" + name2] == 1)
            ].copy()
        elif pairs == "first":
            minPairAS = md_data[
                    (md_data["flag_min_match_" + name1] == 1)
            ].copy()
        elif pairs == "second":
            minPairAS = md_data[
                    (md_data["flag_min_match_" + name2] == 1)
            ].copy()
    # Check if there are any pairs to plot
    if len(minPairAS) == 0:
        return
    minPairAS = minPairAS[
        [
            "transcript_1",
            "transcript_2",
            "flag_alt_exon",
            "flag_alt_donor_acceptor",
            "flag_IR",
            "flag_5_variation",
            "flag_3_variation",
            "flag_no_shared_nt",
            "num_nt_diff",
            "prop_nt_diff",
        ]
    ]
    # Set No Shared NT genes to have np.nan nucleotide difference
    minPairAS["num_nt_diff"] = np.where(
        minPairAS["flag_no_shared_nt"] == 1,
        np.nan,
        minPairAS["num_nt_diff"]
    )
    minPairAS = minPairAS.rename(
        columns={
            "flag_alt_exon": "Alt. Exon",
            "flag_alt_donor_acceptor": "Alt. Donor/Acceptor",
            "flag_IR": "Intron Retention",
            "flag_5_variation": "5' Variation",
            "flag_3_variation": "3' Variation",
            "flag_no_shared_nt": "No Shared NT",
            "num_nt_diff": "# NT Different",
            "prop_nt_diff": "Proportion\nNT Different",
        }
    )
    AScols = [
        "5' Variation",
        "3' Variation",
        "Alt. Donor/Acceptor",
        "Alt. Exon",
        "Intron Retention",
        "No Shared NT",
    ]
    minPairAS[AScols] = minPairAS[AScols].astype(int)
    if reciprocal:
        plot_upset(
            minPairAS.set_index(AScols),
            "Number of Reciprocal Minimum Pairs with AS Categories",
            ["# NT Different", "Proportion\nNT Different"],
        )
        legendText = (
            "Number of reciprocal minimum pairs with the specified types of alternative "
            "splicing between {} and {} indicated by the black dots below the histogram "
            "of pair counts (n = {} total pairs). Pairs with no shared nucleotides are "
            "in the same genes but with nonoverlapping coordinates. Box plots of the "
            "number (blue) and proportion (orange) of nucleotide (NT) differences between "
            "the pairs represented in the histogram.".format(
                name1,
                name2,
                len(minPairAS),
            )
        )
    else:
        if pairs == None:
            plot_upset(
                minPairAS.set_index(AScols),
                "Number of Extra Minimum Pairs with AS Categories",
                ["# NT Different", "Proportion\nNT Different"],
            )
            legendText = (
                "Number of extra minimum pairs, or pairs without reciprocal minimums, "
                "with the specified types of alternative splicing between {} and {} "
                "indicated by the black dots below the histogram of pair counts "
                "(n = {} total pairs). Pairs with no shared nucleotides are "
                "in the same genes but with nonoverlapping coordinates. Box plots of the "
                "number (blue) and proportion (orange) of nucleotide (NT) differences between "
                "the pairs represented in the histogram.".format(
                    name1,
                    name2,
                    len(minPairAS),
                )
            )
        elif pairs == "all":
            plot_upset(
                minPairAS.set_index(AScols),
                "Number of Minimum Pairs in either {} or {} with AS Categories".format(name1, name2),
                ["# NT Different", "Proportion\nNT Different"],
            )
            legendText = (
                "Number of minimum pairs in either {} or {}  "
                "with the specified types of alternative splicing between {} and {} "
                "indicated by the black dots below the histogram of pair counts "
                "(n = {} total pairs). Pairs with no shared nucleotides are "
                "in the same genes but with nonoverlapping coordinates. Box plots of the "
                "number (blue) and proportion (orange) of nucleotide (NT) differences between "
                "the pairs represented in the histogram.".format(
                    name1,
                    name2,
                    name1,
                    name2,
                    len(minPairAS),
                )
            )
        elif pairs == "first":
            plot_upset(
                minPairAS.set_index(AScols),
                "Number of Minimum Pairs in {} with AS Categories".format(name1),
                ["# NT Different", "Proportion\nNT Different"],
            )
            legendText = (
                "Number of minimum pairs in {}  "
                "with the specified types of alternative splicing between {} and {} "
                "indicated by the black dots below the histogram of pair counts "
                "(n = {} total pairs). Pairs with no shared nucleotides are "
                "in the same genes but with nonoverlapping coordinates. Box plots of the "
                "number (blue) and proportion (orange) of nucleotide (NT) differences between "
                "the pairs represented in the histogram.".format(
                    name1,
                    name1,
                    name2,
                    len(minPairAS),
                )
            )
        elif pairs == "second":
            plot_upset(
                minPairAS.set_index(AScols),
                "Number of Minimum Pairs in {} with AS Categories".format(name2),
                ["# NT Different", "Proportion\nNT Different"],
            )
            legendText = (
                "Number of minimum pairs in {}  "
                "with the specified types of alternative splicing between {} and {} "
                "indicated by the black dots below the histogram of pair counts "
                "(n = {} total pairs). Pairs with no shared nucleotides are "
                "in the same genes but with nonoverlapping coordinates. Box plots of the "
                "number (blue) and proportion (orange) of nucleotide (NT) differences between "
                "the pairs represented in the histogram.".format(
                    name2,
                    name1,
                    name2,
                    len(minPairAS),
                )
            )
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        if reciprocal:
            outFile.write(
                r"\b Figure. Reciprocal minimum pair alternative splicing \b0"
                r" \line {}".format(
                    legendText
                )
            )
        else:
            outFile.write(
                r"\b Figure. Extra minimum pair alternative splicing \b0 \line "
                r"{}".format(
                    legendText
                )
            )
        end_rtf(outFile)


def plot_gene_recip_min_AS_upset(md_data, name1, name2, legendOut):
    """
    Upset plot for number of genes with each kind of AS when comparing the
            reciprocally min match pairs of the two datasets with added box plots
            of the number of min match pairs present in the gene and the number
            of transcripts in each dataset
    """
    recipMinPairAS = md_data[md_data["flag_recip_min_match"] == 1][
        [
            "gene_id",
            "flag_alt_exon_recip_min_match",
            "flag_alt_donor_acceptor_recip_min_match",
            "flag_IR_recip_min_match",
            "flag_5_variation_recip_min_match",
            "flag_3_variation_recip_min_match",
            "flag_no_shared_nt_recip_min_match",
            "num_recip_min_match_in_gene",
            "num_transcript_in_gene_" + name1,
            "num_transcript_in_gene_" + name2,
        ]
    ].copy()
    geneRecipMatchAS = (
        recipMinPairAS.groupby("gene_id")
        .agg(
            {
                "flag_alt_exon_recip_min_match": "max",
                "flag_alt_donor_acceptor_recip_min_match": "max",
                "flag_IR_recip_min_match": "max",
                "flag_5_variation_recip_min_match": "max",
                "flag_3_variation_recip_min_match": "max",
                "flag_no_shared_nt_recip_min_match": "max",
                "num_recip_min_match_in_gene": "max",
                "num_transcript_in_gene_" + name1: "max",
                "num_transcript_in_gene_" + name2: "max",
            }
        )
        .reset_index()
    )
    geneFlagCols = [
        "flag_alt_exon_recip_min_match",
        "flag_alt_donor_acceptor_recip_min_match",
        "flag_IR_recip_min_match",
        "flag_5_variation_recip_min_match",
        "flag_3_variation_recip_min_match",
        "flag_no_shared_nt_recip_min_match",
    ]
    geneRecipMatchAS[geneFlagCols] = geneRecipMatchAS[geneFlagCols].astype(bool)
    geneRecipMatchAS = geneRecipMatchAS.rename(
        columns={
            "flag_alt_exon_recip_min_match": "Alt. Exon",
            "flag_alt_donor_acceptor_recip_min_match": "Alt. Donor/Acceptor",
            "flag_IR_recip_min_match": "Intron Retention",
            "flag_5_variation_recip_min_match": "5' Variation",
            "flag_3_variation_recip_min_match": "3' Variation",
            "flag_no_shared_nt_recip_min_match": "No Shared NT",
            "num_recip_min_match_in_gene": "# Recip. Min.\nTranscript Pairs",
            "num_transcript_in_gene_" + name1: "# Transcripts\nin " + name1,
            "num_transcript_in_gene_" + name2: "# Transcripts\nin " + name2,
        }
    )
    AScols = [
        "5' Variation",
        "3' Variation",
        "Alt. Donor/Acceptor",
        "Alt. Exon",
        "Intron Retention",
        "No Shared NT",
    ]
    plot_upset(
        geneRecipMatchAS.set_index(AScols),
        "",
        [
            "# Recip. Min.\nTranscript Pairs",
            "# Transcripts\nin " + name1,
            "# Transcripts\nin " + name2,
        ],
    )
    legendText = (
        "Number of genes with the specified types of alternative splicing in only "
        "reciprocal minimum pairs between {} and {} indicated by the black dots "
        "below the histogram of gene counts (n = {} genes with {} reciprocal "
        "minimum pairs). Box plots represent the number of reciprocal minimum "
        "pairs (blue) and the number of transcripts in {} (orange) and {} "
        "(green). Genes with \"No Shared NT\" have a pair of "
        "transcripts with nonoverlapping coordinates.".format(
            name1,
            name2,
            len(geneRecipMatchAS),
            len(md_data[md_data["flag_recip_min_match"] == 1]),
            name1,
            name2,
        )
    )
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Observed alternative splicing in genes among "
            r"reciprocal minimum pairs of transcripts \b0 \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def plot_gene_AS_upset(td_data, legendOut):
    """
    Upset plot for number of genes with each kind of AS when comparing all
            transcript pairs within each gene of a single transcriptome with added box plot
            of the number transcripts per gene
    """
    pairAS = td_data[
        [
            "gene_id",
            "transcript_1",
            "transcript_2",
            "flag_alt_exon",
            "flag_alt_donor_acceptor",
            "flag_IR",
            "flag_5_variation",
            "flag_3_variation",
            "flag_no_shared_nt",
        ]
    ].copy()
    # Get AS flags at the gene level
    geneAS = (
        pairAS.groupby("gene_id")
        .agg(
            {
                "flag_alt_exon": "max",
                "flag_alt_donor_acceptor": "max",
                "flag_IR": "max",
                "flag_5_variation": "max",
                "flag_3_variation": "max",
                "flag_no_shared_nt": "max",
            }
        )
        .reset_index()
    )
    # Get the number of transcripts per gene by getting the unique list
    #   of transcript_1 and transcript_2
    transcriptPerGene = (
        pd.concat(
            [
                pairAS[["gene_id", "transcript_1"]].rename(
                    columns={"transcript_1": "transcript_id"}
                ),
                pairAS[["gene_id", "transcript_2"]].rename(
                    columns={"transcript_2": "transcript_id"}
                ),
            ]
        )
        .groupby("gene_id")["transcript_id"]
        .nunique()
        .reset_index()
        .rename(columns={"transcript_id": "num_transcript_per_gene"})
    )
    # Merge transcripts per gene with AS flags
    mergeASxcrptPerGene = pd.merge(
        geneAS, transcriptPerGene, how="outer", on="gene_id", indicator="merge_check"
    )
    # Set AS flags to boolean values and rename
    geneFlagCols = [
        "gene_id",
        "flag_alt_exon",
        "flag_alt_donor_acceptor",
        "flag_IR",
        "flag_5_variation",
        "flag_3_variation",
        "flag_no_shared_nt",
    ]
    mergeASxcrptPerGene[geneFlagCols] = mergeASxcrptPerGene[geneFlagCols].astype(bool)
    mergeASxcrptPerGene = mergeASxcrptPerGene.rename(
        columns={
            "flag_alt_exon": "Alt. Exon",
            "flag_alt_donor_acceptor": "Alt. Donor/Acceptor",
            "flag_IR": "Intron Retention",
            "flag_5_variation": "5' Variation",
            "flag_3_variation": "3' Variation",
            "flag_no_shared_nt": "No Shared NT",
            "num_transcript_per_gene": "# Transcripts\nPer Gene",
            "num_nt_diff": "Avg #\nNT Different",
            "prop_nt_diff": "Avg Proportion\nNT Different",
        }
    )
    AScols = [
        "5' Variation",
        "3' Variation",
        "Alt. Donor/Acceptor",
        "Alt. Exon",
        "Intron Retention",
        "No Shared NT",
    ]
    plot_upset(
        mergeASxcrptPerGene.set_index(AScols),
        "",
        [
            "# Transcripts\nPer Gene",
        ],
    )
    legendText = (
        "Number of genes with the specified types of alternative splicing indicated "
        "by the black dots below the histogram of gene counts (n = {} multi-transcript genes). "
        "The box plot represent the number of transcripts per gene (blue). "
        "Genes with \"No Shared NT\" have a pair of transcripts with "
        "nonoverlapping coordinates.".format(
            len(mergeASxcrptPerGene)
        )
    )
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Alternative splicing in genes \b0 \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def plot_pair_AS_upset_nt_box(td_data, legendOut):
    """
    Upset plot for the number of transcript pairs with each kind of AS
            for a single transcriptome with added box plots
            of the number of nt different and the proportion of nt different
    """
    pairAS = td_data[
        [
            "gene_id",
            "transcript_1",
            "transcript_2",
            "flag_alt_exon",
            "flag_alt_donor_acceptor",
            "flag_IR",
            "flag_5_variation",
            "flag_3_variation",
            "flag_no_shared_nt",
            "num_nt_diff",
            "prop_nt_diff",
        ]
    ].copy()
    # Ensure number of nt different are int and float values
    pairAS["num_nt_diff"] = pairAS["num_nt_diff"].astype(int)
    pairAS["prop_nt_diff"] = pairAS["prop_nt_diff"].astype(float)
    # Set No Shared NT pairs to have np.nan nucleotide difference
    pairAS["num_nt_diff"] = np.where(
        pairAS["flag_no_shared_nt"] == 1,
        np.nan,
        pairAS["num_nt_diff"]
    )
    # Make AS flags boolean values and rename
    xcrptFlagCols = [
        "transcript_1",
        "transcript_2",
        "flag_alt_exon",
        "flag_alt_donor_acceptor",
        "flag_IR",
        "flag_5_variation",
        "flag_3_variation",
        "flag_no_shared_nt",
    ]
    pairAS[xcrptFlagCols] = pairAS[xcrptFlagCols].astype(bool)
    pairAS = pairAS.rename(
        columns={
            "flag_alt_exon": "Alt. Exon",
            "flag_alt_donor_acceptor": "Alt. Donor/Acceptor",
            "flag_IR": "Intron Retention",
            "flag_5_variation": "5' Variation",
            "flag_3_variation": "3' Variation",
            "flag_no_shared_nt": "No Shared NT",
            "num_nt_diff": "# NT Different",
            "prop_nt_diff": "Proportion\nNT Different",
        }
    )
    AScols = [
        "5' Variation",
        "3' Variation",
        "Alt. Donor/Acceptor",
        "Alt. Exon",
        "Intron Retention",
        "No Shared NT",
    ]
    plot_upset(
        pairAS.set_index(AScols),
        "",
        [
            "# NT Different",
            "Proportion\nNT Different",
        ],
    )
    legendText = (
        "Number of transcript pairs with the specified types of alternative "
        "splicing indicated by the black dots below the histogram of transcript "
        "pair counts (n = {} transcript pairs, in {} multi-transcript genes). "
        "Box plots represent the number (blue) and proportion (orange) of "
        "nucleotides different between the pair. Transcript pairs with "
        "\"No Shared NT\" have nonoverlapping coordinates.".format(
            len(pairAS),
            pairAS["gene_id"].nunique()
        )
    )
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Alternative splicing between pairs of transcripts \b0"
            r" \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def plot_upset(df, title, boxCols=None):
    """
    Plot an UpSet plot given a properly formatted dataframe (multi-indexed with boolean values)
    """
    upset = UpSet(
        df,
        subset_size="count",
        show_counts=True,
        sort_by="degree",
        sort_categories_by=None,
    )
    if boxCols is not None:
        if type(boxCols) != list:
            boxCols = list(boxCols)
        for col in boxCols:
            upset.add_catplot(
                value=col,
                kind="box",
                elements=4,
                showfliers=False,
                color=sns.color_palette("colorblind", 15).as_hex()[boxCols.index(col)],
            )
    upset.plot()
    plt.subplots_adjust(right=1.00001)
    plt.suptitle(title)


def plot_gene_avg_nt_diff_pairs(md_data, name1, name2, legendOut, zoomMean=False):
    """
    Plot average number of nt different in recip. min match pairs against
            avg number of nt different in min pairs that are extra
            (either in the greater sets of in set that do not match) with the
            option to zoom in to the  mean value of non-zero nucleotide differences
            (max of the two means from recip. min pairs and extras)
    """
    minPairNT = md_data[
        [
            "gene_id",
            "transcript_in_gene",
            "recip_min_pair_in_gene",
            "flag_recip_min_match",
            "flag_min_match_" + name1,
            "flag_min_match_" + name2,
            "num_nt_diff",
            "prop_nt_diff",
        ]
    ].copy()
    # Ensure number of nt different are int and float values
    minPairNT["num_nt_diff"] = minPairNT["num_nt_diff"].astype(int)
    minPairNT["prop_nt_diff"] = minPairNT["prop_nt_diff"].astype(float)
    minPairNT["num_nt_diff_recip_min"] = np.where(
        minPairNT["flag_recip_min_match"] == 1, minPairNT["num_nt_diff"], 0
    )
    minPairNT["num_nt_diff_extra"] = np.where(
        (minPairNT["flag_recip_min_match"] == 0)
        & (
            minPairNT["flag_min_match_" + name1] + minPairNT["flag_min_match_" + name2]
            == 1
        ),
        minPairNT["num_nt_diff"],
        0,
    )
    minPairNTGene = (
        minPairNT.groupby("gene_id")
        .agg(
            {
                "transcript_in_gene": "first",
                "recip_min_pair_in_gene": "first",
                "num_nt_diff_recip_min": "mean",
                "num_nt_diff_extra": "mean",
            }
        )
        .reset_index()
    )
    recipConditions = [
        (minPairNTGene["transcript_in_gene"] == "match")
        & (minPairNTGene["recip_min_pair_in_gene"] == "reciprocal_pairs"),
        (minPairNTGene["transcript_in_gene"] == "match")
        & (minPairNTGene["recip_min_pair_in_gene"] == "partial_reciprocal_pairs"),
        (minPairNTGene["transcript_in_gene"] == "match")
        & (minPairNTGene["recip_min_pair_in_gene"] == "no_reciprocal_pairs"),
        (minPairNTGene["transcript_in_gene"] == name1 + "_greater")
        & (minPairNTGene["recip_min_pair_in_gene"] == "reciprocal_pairs"),
        (minPairNTGene["transcript_in_gene"] == name1 + "_greater")
        & (minPairNTGene["recip_min_pair_in_gene"] == "partial_reciprocal_pairs"),
        (minPairNTGene["transcript_in_gene"] == name1 + "_greater")
        & (minPairNTGene["recip_min_pair_in_gene"] == "no_reciprocal_pairs"),
        (minPairNTGene["transcript_in_gene"] == name2 + "_greater")
        & (minPairNTGene["recip_min_pair_in_gene"] == "reciprocal_pairs"),
        (minPairNTGene["transcript_in_gene"] == name2 + "_greater")
        & (minPairNTGene["recip_min_pair_in_gene"] == "partial_reciprocal_pairs"),
        (minPairNTGene["transcript_in_gene"] == name2 + "_greater")
        & (minPairNTGene["recip_min_pair_in_gene"] == "no_reciprocal_pairs"),
    ]
    recipChoices = [
        "Match:Reciprocal Pairs",
        "Match:Partial Reciprocal Pairs",
        "Match:No Reciprocal Pairs",
        name1 + " Greater:Reciprocal Pairs",
        name1 + " Greater:Partial Reciprocal Pairs",
        name1 + " Greater:No Reciprocal Pairs",
        name2 + " Greater:Reciprocal Pairs",
        name2 + " Greater:Partial Reciprocal Pairs",
        name2 + " Greater:No Reciprocal Pairs",
    ]
    minPairNTGene["Category"] = np.select(recipConditions, recipChoices, "NC")
    minPairGeneOrder = {
        "Match:Reciprocal Pairs": 0,
        "Match:Partial Reciprocal Pairs": 1,
        "Match:No Reciprocal Pairs": 2,
        name1 + " Greater:Reciprocal Pairs": 3,
        name1 + " Greater:Partial Reciprocal Pairs": 4,
        name1 + " Greater:No Reciprocal Pairs": 5,
        name2 + " Greater:Reciprocal Pairs": 6,
        name2 + " Greater:Partial Reciprocal Pairs": 7,
        name2 + " Greater:No Reciprocal Pairs": 8,
    }
    minPairNTGene["Category"] = pd.Categorical(
        minPairNTGene["Category"],
        categories=sorted(minPairGeneOrder, key=minPairGeneOrder.get),
        ordered=True,
    )
    minPairNTGene = minPairNTGene.sort_values("Category")
    colorPalleteRecip = {
        "Match:Reciprocal Pairs": "#0173b2",
        "Match:Partial Reciprocal Pairs": "#56b4e9",
        "Match:No Reciprocal Pairs": "#029e73",
        name1 + " Greater:Reciprocal Pairs": "#d55e00",
        name1 + " Greater:Partial Reciprocal Pairs": "#de8f05",
        name1 + " Greater:No Reciprocal Pairs": "#ece133",
        name2 + " Greater:Reciprocal Pairs": "#949494",
        name2 + " Greater:Partial Reciprocal Pairs": "#fbafe4",
        name2 + " Greater:No Reciprocal Pairs": "#ca9161",
    }
    if not zoomMean:
        limit = (
            max(
                minPairNTGene[
                        ["num_nt_diff_recip_min", "num_nt_diff_extra"]
                        ].max().max(),
                1
            )
        )
    else:
        try:
            limit = round(
                np.nanmax(
                    [
                        minPairNTGene[minPairNTGene["num_nt_diff_recip_min"] > 0][
                            "num_nt_diff_recip_min"
                        ].mean(),
                        minPairNTGene[minPairNTGene["num_nt_diff_extra"] > 0][
                            "num_nt_diff_extra"
                        ].mean(),
                    ]
                )
            )
        except ValueError:
            limit = 1
    plt.figure(figsize=(8.45, 5))
    sns.scatterplot(
        data=minPairNTGene,
        x="num_nt_diff_recip_min",
        y="num_nt_diff_extra",
        hue="Category",
        palette=colorPalleteRecip,
    )
    plt.xlim(0, limit)
    plt.ylim(0, limit)
    plt.xlabel("Avg. # NT Different in Recip. Min. Pairs")
    plt.ylabel("Avg. # NT Different in Min. Pairs of Extras")
    plt.legend(bbox_to_anchor=(1.04, 1), loc="upper left")
    plt.plot((0, limit), (0, limit), color="r")
    legendText = (
        "Each point represents a gene plotted by the average number of nucleotides "
        "different between reciprocal minimum pairs (X-axis) and minimum pairs "
        "of extras without a reciprocal minimum pair (Y-axis). Genes are colored "
        "by categorizations of the comparison between {} and {}.".format(
            name1,
            name2,
        )
    )
    plt.tight_layout()
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Number of nucleotides different between reciprocal "
            r"and nonreciprocal minimum pairs \b0 \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def plot_complexity_box(geneDF, transcriptDF, outdir, legendOut):
    """
    Plot box plots of complexity measures:
        1) Transcripts per gene
        2) Unique exons per gene
        3) Exons per transcript
    """
    #    fig = plt.figure(figsize=(6,12))
    fig = plt.figure(figsize=(12, 5))
    #    axTop = plt.subplot2grid((3,1),(0,0),fig=fig)
    axTop = plt.subplot2grid((1, 3), (0, 0), fig=fig)
    axTop.set_xlabel("")
    #    axTop.set_ylabel("Number of Transcripts Per Gene")
    axTop.set_title("Number of Transcripts Per Gene")
    axTop.set_ylabel("")
    #    axMid = plt.subplot2grid((3,1),(1,0),fig=fig)
    axMid = plt.subplot2grid((1, 3), (0, 1), fig=fig)
    axMid.set_xlabel("")
    #    axMid.set_ylabel("Number of Unique Exons Per Gene")
    axMid.set_title("Number of Unique Exons Per Gene")
    axMid.set_ylabel("")
    #    axBottom = plt.subplot2grid((3,1),(2,0),fig=fig)
    axBottom = plt.subplot2grid((1, 3), (0, 2), fig=fig)
    axBottom.set_xlabel("")
    #    axBottom.set_ylabel("Number of Exons Per Transcript")
    axBottom.set_title("Number of Exons Per Transcript")
    axBottom.set_ylabel("")
    bplotTop = geneDF.set_index("gene_id")[["num_transcript"]].boxplot(
        ax=axTop, notch=True, patch_artist=True, return_type="dict"
    )
    axTop.set_xticks([])
    bplotMid = geneDF.set_index("gene_id")[["num_uniq_exon"]].boxplot(
        ax=axMid, notch=True, patch_artist=True, return_type="dict"
    )
    axMid.set_xticks([])
    bplotBottom = transcriptDF.set_index("transcript_id")[["num_exon"]].boxplot(
        ax=axBottom, notch=True, patch_artist=True, return_type="dict"
    )
    axBottom.set_xticks([])
    for bplot in (bplotTop, bplotMid, bplotBottom):
        for patch in bplot["boxes"]:
            patch.set_facecolor("turquoise")
            patch.set_edgecolor("black")
        for median in bplot["medians"]:
            median.set(color="red")
        for whisker in bplot["whiskers"]:
            whisker.set(color="black")
    legendText = (
        "Distributions of transcriptome complexity measures including transcripts "
        "per gene (median = {}, mean = {}), unique exons per gene "
        "(median = {}, mean = {}), and exons per transcript "
        "(median = {}, mean = {}). Total number of genes = {}. "
        "Total number of transcripts = {}.".format(
            geneDF["num_transcript"].median(),
            geneDF["num_transcript"].mean(),
            geneDF["num_uniq_exon"].median(),
            geneDF["num_uniq_exon"].mean(),
            transcriptDF["num_exon"].median(),
            transcriptDF["num_exon"].mean(),
            len(geneDF),
            int(geneDF["num_transcript"].sum()),
        )
    )
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Transcriptome complexity \b0 \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def plot_complexity_violin(geneDF, transcriptDF, outdir, legendOut):
    """
    Plot violin plots of complexity measures:
        1) Transcripts per gene
        2) Unique exons per gene
        3) Exons per transcript
    """
    sns.set(style="whitegrid")
    fig = plt.figure(figsize=(12, 5))
    axLeft = plt.subplot2grid((1, 3), (0, 0), fig=fig)
    axLeft.set_title("Number of Transcripts Per Gene")
    axMid = plt.subplot2grid((1, 3), (0, 1), fig=fig)
    axMid.set_title("Number of Unique Exons Per Gene")
    axRight = plt.subplot2grid((1, 3), (0, 2), fig=fig)
    axRight.set_title("Number of Exons Per Transcript")
    sns.violinplot(y="num_transcript", data=geneDF.set_index("gene_id"), ax=axLeft)
    sns.violinplot(y="num_uniq_exon", data=geneDF.set_index("gene_id"), ax=axMid)
    sns.violinplot(
        y="num_exon", data=transcriptDF.set_index("transcript_id"), ax=axRight
    )
    axLeft.set_xlabel("")
    axLeft.set_ylabel("")
    axMid.set_xlabel("")
    axMid.set_ylabel("")
    axRight.set_xlabel("")
    axRight.set_ylabel("")
    plt.tight_layout()
    legendText = (
        "Distributions of transcriptome complexity measures including transcripts "
        "per gene (median = {}, mean = {}), unique exons per gene (median = {}, mean = {}), "
        "and exons per transcript (median = {}, mean = {}). "
        "Total genes described = {}.".format(
            geneDF["num_transcript"].median(),
            geneDF["num_transcript"].mean(),
            geneDF["num_uniq_exon"].median(),
            geneDF["num_uniq_exon"].mean(),
            transcriptDF["num_exon"].median(),
            transcriptDF["num_exon"].mean(),
            len(geneDF),
        )
    )
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Transcriptome complexity \b0 \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def plot_gene_prop_nt_variablility(ef_data, legendOut, multitranscript=False):
    """
    Plot distribution of proportion of nucleotide variability across all genes or
    only multitranscript genes
    """
    # Get lengths of exon fragments
    ef_data["num_nt_varying"] = np.where(
        ef_data["ea_annotation_frequency"] != "constitutive",
        ef_data["ef_end"] - ef_data["ef_start"],
        0,
    )
    ef_data["num_nt_total"] = ef_data["ef_end"] - ef_data["ef_start"]
    # For each gene, get the number of varying nucleotides (not constitutive)
    ef_gene_data = (
        ef_data.groupby("gene_id")[["num_nt_varying", "num_nt_total"]]
        .sum()
        .reset_index()
    )
    # Get propotion of varying nucleotides
    ef_gene_data["prop_varying_nt"] = (
        ef_gene_data["num_nt_varying"] / ef_gene_data["num_nt_total"]
    )

    # Check if outputting all genes or only multitranscript genes
    if multitranscript:
        ef_data_split = split_column_by_sep(
            ef_data,
            col_name="ef_transcript_ids",
            sep="|"
        )
        xcrpt_per_gene = (
            ef_data_split.groupby("gene_id")["ef_transcript_ids"]
            .nunique()
            .reset_index()
        )
        multi_xcrpt_gene = xcrpt_per_gene[xcrpt_per_gene["ef_transcript_ids"] > 1]
        ef_gene_data = ef_gene_data[
            ef_gene_data["gene_id"].isin(multi_xcrpt_gene["gene_id"])
        ]
        legendText = (
            "Distribution of the proportion of variable "
            "nucleotides across multi-transcript genes (n = {}).".format(
                len(ef_gene_data)
            )
        )
    else:
        legendText = (
            "Distribution of the proportion of variable "
            "nucleotides across all genes (n = {}). "
            "Single transcript genes are assigned a proportion of 1.".format(
                len(ef_gene_data)
            )
        )
    # Plot density
    sns.distplot(
            ef_gene_data["prop_varying_nt"],
            hist=False
    )
    plt.xlim(0, 1)
    plt.ylabel("Density")
    plt.xlabel("Proportion of Variable Nucleotides")
    plt.tight_layout()
    with open(legendOut, "w") as outFile:
        start_rtf(outFile)
        outFile.write(
            r"\b Figure. Transcriptome complexity \b0 \line {}".format(
                legendText
            )
        )
        end_rtf(outFile)


def split_column_by_sep(df, col_name=None, sep=None, sort_list=None):
    """
    Split variable by some character like '|' or ',' and
    keep all other values the same
    """
    if col_name is None:
        col_name = "transcript_id"
    if sep is None:
        sep = "|"
    splitList = df[col_name].str.split(sep).apply(pd.Series, 1).stack()
    splitList.index = splitList.index.droplevel(-1)
    tempDF = df.copy()
    del tempDF[col_name]
    splitDF = tempDF.join(splitList.rename(col_name))
    if sort_list is not None:
        splitDF = splitDF.sort_values(by=sort_list)
    del (tempDF, splitList)
    return splitDF


def get_value_count(df, col, value):
    """
    Do value_count function of column col on dataframe df
    Get count of value (an element in col), if value is not present then return 0
    """
    df2 = df[col].value_counts()
    if value in df2.index:
        value = df2[value]
    else:
        value = 0
    return value


def start_rtf(outFile):
    """
    Open RTF file
    """
    outFile.write(
        r"{\rtf1\ansi\ansicpg1252\deff0\deflang1033{\fonttbl{\f0\fswiss\fcharset0 Arial;}}"
    )

def get_citation():
    return (
        r" \line \line 1. Nanni, A., Titus-McQuillan, J., Moskalenko, O., "
        r"Pardo-Palacios, F., Liu, Z., Conesa, A., Rogers, R. L., & McIntyre, "
        "L. M. (2021). The evolution of splicing: transcriptome complexity "
        "and transcript distances implemented in TranD. \i bioRxiv\i0, "
        "2021.2009.2028.462251. https://doi.org/10.1101/2021.09.28.462251. "
        r"https://github.com/McIntyre-Lab/TranD."
    )

def end_rtf(outFile):
    """
    Close RTF file
    """
    outFile.write(
        r" Transcriptome analyses performed by TranD [1]. {}"
        r" \line \line \line \i Disclaimer:  While automated captions of "
        r"TranD have been carefully constructed, users are advised to verify "
        r"caption contents before use and report any errors to the TranD "
        r"github.\i0 ".format(
            get_citation()
        )
    )
    outFile.write(r"}\n\x00")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 24 14:38:56 2021

@author: Adalena Nanni <adalena.nanni@ufl.edu>, Oleksandr Moskalenko <om@rc.ufl.edu
"""

import pandas as pd
import numpy as np
from loguru import logger
from .bedtools import prep_bed_for_ea

# CONFIGURATION
# Duplicate to prevent a circular import
jct_df_cols = ['gene_id', 'transcript_id', 'coords']


def create_junction_catalog_str(gene, tx, tx_data):
    """Create a junction catalog for a transcript"""
    junctions = []
    id = 0
    tx_data_df = pd.DataFrame(tx_data, columns=['chrom', 'start', 'end', 'name',
                              'score', 'strand'])
    tx_data_df['start'] = tx_data_df['start'].astype(int)
    tx_data_df['end'] = tx_data_df['end'].astype(int)
    tx_data_df = tx_data_df.sort_values('start')
    for index, e in tx_data_df.iterrows():
        id += 1
        if id == 1:
            left_end = int(e.end) - 10
            continue
        right_start = int(e.start + 10)
        jct_coords = f"{e.chrom}:{left_end}:{right_start}:{e.strand}"
        junctions.append([gene, tx, jct_coords])
        left_end = int(e.end) - 10
    return junctions


def consolidate_junctions(
    bed_gene_data, pre_consol_jct_df, prefix="tr"
):
    gene_id = bed_gene_data["gene_id"]
    logger.debug("Consolidation of {}", gene_id)

    # Set output data dataframe
    consol_gene_cols = ["seqname", "start", "end", "strand", "gene_id", "transcript_id"]
    consol_gene = pd.DataFrame(columns=consol_gene_cols)

    # Check if single transcript, if yes then it cannot be consolidated
    if len(bed_gene_data["transcript_list"]) == 1:
        logger.debug("Gene {} has a single transcript.", gene_id)
        xcrpt_all = pd.DataFrame(
            bed_gene_data["transcript_list"], columns=["transcript_id"]
        )
        xcrpt_all["gene_id"] = gene_id
        xcrpt_all["chr"] = bed_gene_data[xcrpt_all["transcript_id"][0]][0][0]
        xcrpt_all["strand"] = bed_gene_data[xcrpt_all["transcript_id"][0]][0][5]
        xcrpt_all["consolidation_transcript_id"] = prefix + "_" + gene_id + "_1"
        key_gene = xcrpt_all[
            ["gene_id", "transcript_id", "consolidation_transcript_id"]
        ]
        # Format output data
        for consol_transcript, transcript in (
            key_gene.groupby("consolidation_transcript_id")["transcript_id"]
            .first()
            .iteritems()
        ):
            for row in bed_gene_data[transcript]:
                consol_gene = pd.concat(
                    [
                        consol_gene,
                        pd.DataFrame(
                            [
                                [
                                    row[0],
                                    int(row[1]) + 1,
                                    int(row[2]),
                                    row[5],
                                    gene_id,
                                    consol_transcript,
                                ]
                            ],
                            columns=consol_gene.columns,
                        ),
                    ],
                    ignore_index=True,
                )
                consol_gene = consol_gene.sort_values(
                    by=["transcript_id", "start"]
                ).reset_index(drop=True)

    # Check if more than one transcript present and all are monoexon (no junctions present)
    elif len(bed_gene_data["transcript_list"]) > 1 and len(pre_consol_jct_df) == 0:
        logger.debug(
            "Gene {} has all monoexon transcripts (more than 1 transcript present).",
            gene_id,
        )
        # Get transcript list and start/end of transcripts
        xcrpt_all = pd.DataFrame(
            bed_gene_data["transcript_list"], columns=["transcript_id"]
        )
        xcrpt_all["gene_id"] = gene_id
        xcrpt_all["chr"] = bed_gene_data[xcrpt_all["transcript_id"][0]][0][0]
        xcrpt_all["strand"] = bed_gene_data[xcrpt_all["transcript_id"][0]][0][5]
        xcrpt_all["start"] = xcrpt_all["transcript_id"].map(
            lambda xcrpt: min([int(x[1]) + 1 for x in bed_gene_data[xcrpt]])
        )
        xcrpt_all["end"] = xcrpt_all["transcript_id"].map(
            lambda xcrpt: max([int(x[2]) for x in bed_gene_data[xcrpt]])
        )
        # Loop over coordinates to make bins of overlapping monoexon transcripts
        xcrpt_all = xcrpt_all.sort_values(by=["start", "end"]).reset_index(drop=True)
        bin_num = 1
        bin_start = xcrpt_all["start"][0]
        bin_end = xcrpt_all["end"][0]
        xcrpt_all["monoexon_group"] = 0
        for index, row in xcrpt_all.iterrows():
            if row["start"] >= bin_start and row["start"] <= bin_end:
                xcrpt_all.loc[index, "monoexon_group"] = bin_num
                if row["end"] >= bin_end:
                    bin_end = row["end"]
            else:
                bin_num += 1
                bin_start = row["start"]
                bin_end = row["end"]
                xcrpt_all.loc[index, "monoexon_group"] = bin_num
        # Group by assigned bins to get min start and max end
        xcrpt_all["min_group_start"] = xcrpt_all.groupby("monoexon_group")[
            "start"
        ].transform("min")
        xcrpt_all["max_group_end"] = xcrpt_all.groupby("monoexon_group")[
            "end"
        ].transform("max")
        xcrpt_all["consolidation_transcript_id"] = (
            prefix
            + "_"
            + gene_id
            + "_"
            + xcrpt_all["monoexon_group"].astype(int).map(str)
        )
        key_gene = xcrpt_all[
            ["gene_id", "transcript_id", "consolidation_transcript_id"]
        ]
        # Format output data
        consol_xcrpt = (
            xcrpt_all.groupby("monoexon_group")[
                ["gene_id", "chr", "strand", "min_group_start", "max_group_end"]
            ]
            .first()
            .reset_index()
        )
        consol_xcrpt["consolidation_transcript_id"] = (
            prefix
            + "_"
            + gene_id
            + "_"
            + consol_xcrpt["monoexon_group"].astype(int).map(str)
        )
        consol_gene = consol_xcrpt[
            [
                "chr",
                "min_group_start",
                "max_group_end",
                "strand",
                "gene_id",
                "consolidation_transcript_id",
            ]
        ]
        consol_gene.columns = consol_gene_cols
    # Gene contains at least one multiexon transcript and more than one transcript total
    else:
        logger.debug(
            "Gene {} contains at least one multiexon transcript (more than 1 transcript present).",
            gene_id,
        )
        # Get junction string for each transcript in the gene
        pre_consol_jct_df = pre_consol_jct_df.sort_values(
            ["transcript_id", "coords"]
        ).reset_index(drop=True)
        collapse_pre_jct = (
            pre_consol_jct_df.groupby("transcript_id")
            .apply(func=lambda x: "|".join(x["coords"]))
            .reset_index()
            .rename(columns={0: "junctionID_order"})
            .sort_values(["junctionID_order"])
            .reset_index(drop=True)
        )

        # Merge transcript-level junction (multiexon transcripts only) and
        #     transcript_id values (all transcripts including multi-exon) variables to get
        #     mono-exon transcripts
        xcrpt_all = pd.DataFrame(
            bed_gene_data["transcript_list"], columns=["transcript_id"]
        )
        xcrpt_all["gene_id"] = gene_id
        xcrpt_all["chr"] = bed_gene_data[xcrpt_all["transcript_id"][0]][0][0]
        xcrpt_all["strand"] = bed_gene_data[xcrpt_all["transcript_id"][0]][0][5]
        xcrpt_jct_w_mono = pd.merge(
            xcrpt_all, collapse_pre_jct, how="outer", on="transcript_id", validate="1:1"
        )
        xcrpt_jct_w_mono["start"] = xcrpt_jct_w_mono["transcript_id"].map(
            lambda xcrpt: min([int(x[1]) + 1 for x in bed_gene_data[xcrpt]])
        )
        xcrpt_jct_w_mono["end"] = xcrpt_jct_w_mono["transcript_id"].map(
            lambda xcrpt: max([int(x[2]) for x in bed_gene_data[xcrpt]])
        )

        # Flag monoexon transcripts (no junctionID_order present)
        xcrpt_jct_w_mono["flag_monoexon_transcript"] = np.where(
            xcrpt_jct_w_mono["junctionID_order"].isna(), 1, 0
        )

        # If present, get monoexon junctionID_order values
        #   using group number for overlapping monoexon transcripts
        if xcrpt_jct_w_mono["flag_monoexon_transcript"].sum() > 0:
            # Loop over coordinates to make bins of overlapping monoexon transcripts
            xcrpt_jct_w_mono = xcrpt_jct_w_mono.sort_values(
                by=["start", "end"]
            ).reset_index(drop=True)
            bin_num = 1
            bin_start = xcrpt_jct_w_mono[
                xcrpt_jct_w_mono["flag_monoexon_transcript"] == 1
            ]["start"].values[0]
            bin_end = xcrpt_jct_w_mono[
                xcrpt_jct_w_mono["flag_monoexon_transcript"] == 1
            ]["end"].values[0]
            xcrpt_jct_w_mono["monoexon_group"] = 0
            for index, row in xcrpt_jct_w_mono[
                xcrpt_jct_w_mono["flag_monoexon_transcript"] == 1
            ].iterrows():
                if row["start"] >= bin_start and row["start"] <= bin_end:
                    xcrpt_jct_w_mono.loc[index, "monoexon_group"] = bin_num
                    if row["end"] >= bin_end:
                        bin_end = row["end"]
                else:
                    bin_num += 1
                    bin_start = row["start"]
                    bin_end = row["end"]
                    xcrpt_jct_w_mono.loc[index, "monoexon_group"] = bin_num
            logger.debug(
                "Gene {} has {} monoexon transcript(s) that occupy {} nonoverlapping "
                "genomic region(s).",
                gene_id,
                xcrpt_jct_w_mono["flag_monoexon_transcript"].sum(),
                bin_num,
            )
            # Set junctionID_order to monoexon group number
            xcrpt_jct_w_mono = xcrpt_jct_w_mono.fillna(-1)
            xcrpt_jct_w_mono.loc[
                xcrpt_jct_w_mono["junctionID_order"] == -1, "junctionID_order"
            ] = xcrpt_jct_w_mono["monoexon_group"]

        # Make groups of transcripts with same junctionID_order
        # Group names will be piped list of transcript IDs that share the same junctions
        xcrpt_jct_w_mono["transcriptID_cat"] = xcrpt_jct_w_mono.groupby(
            ["junctionID_order"]
        )["transcript_id"].transform(func=lambda x: "|".join(x))

        # Get minimum start and maximum end for longest 5' and 3' ends within each group
        #     with the same junctionID_order
        xcrpt_jct_w_mono["transcript_length"] = xcrpt_jct_w_mono["end"].astype(
            int
        ) - xcrpt_jct_w_mono["start"].astype(int)
        xcrpt_jct_w_mono["min_group_start"] = xcrpt_jct_w_mono.groupby(
            "junctionID_order"
        )["start"].transform("min")
        xcrpt_jct_w_mono["max_group_end"] = xcrpt_jct_w_mono.groupby(
            "junctionID_order"
        )["end"].transform("max")

        # Select the longest transcript for each junctionID_order
        longest_df = xcrpt_jct_w_mono.loc[
            xcrpt_jct_w_mono.groupby("junctionID_order")["transcript_length"].idxmax()
        ]

        # Generate new transcript_id for each transcript:
        #     [prefix]_[gene_id]_# where #={1,...,n} for all n transcripts in the gene
        # First sort transcript_length
        longest_df["consol_transcript_length"] = (
            longest_df["max_group_end"] - longest_df["min_group_start"]
        )
        longest_df = longest_df.sort_values(
            by="consol_transcript_length", ascending=False
        ).reset_index(drop=True)
        longest_df["transcript_rank_in_gene"] = longest_df[
            "consol_transcript_length"
        ].rank(method="first")
        longest_df["consolidation_transcript_id"] = (
            prefix
            + "_"
            + gene_id
            + "_"
            + longest_df["transcript_rank_in_gene"].astype(int).map(str)
        )

        # Flag transcripts where min start does not match the longest representative start

        longest_df["flag_not_min_start"] = np.where(
            longest_df["start"] != longest_df["min_group_start"], 1, 0
        )
        # Flag transcripts where the max end does not match the end of the longest representative
        longest_df["flag_not_max_end"] = np.where(
            longest_df["end"] != longest_df["max_group_end"], 1, 0
        )

        # Set consolidated start and end
        longest_df["consol_start"] = np.where(
            longest_df["flag_not_min_start"] == 1,
            longest_df["min_group_start"],
            longest_df["start"],
        )
        longest_df["consol_end"] = np.where(
            longest_df["flag_not_max_end"] == 1,
            longest_df["max_group_end"],
            longest_df["end"],
        )

        # Make key for gene_id 2 transcript_id 2 consolidation_transcript_id
        key_gene = pd.merge(
            xcrpt_jct_w_mono,
            longest_df[["consolidation_transcript_id", "junctionID_order"]],
            how="left",
            on="junctionID_order",
            validate="m:1",
        )
        key_gene["gene_id"] = gene_id
        key_gene = key_gene[["gene_id", "transcript_id", "consolidation_transcript_id"]]

        # Format output data by setting first and last exon coordinates based on longest first and
        # last exons of the groups Internal exons are the same for all transcripts in the group so
        # coordinates are kept for those.
        for consol_transcript, transcript in (
            key_gene.groupby("consolidation_transcript_id")["transcript_id"]
            .first()
            .iteritems()
        ):
            num = 0
            # Make start and end values integers for proper sorting
            for val in range(0, len(bed_gene_data[transcript])):
                tupleList = list(bed_gene_data[transcript][val])
                tupleList[1] = int(tupleList[1])
                tupleList[2] = int(tupleList[2])
                bed_gene_data[transcript][val] = tuple(tupleList)
            # Check if monoexon to use the longest start and end
            if len(bed_gene_data[transcript]) == 1:
                consol_gene = pd.concat(
                        [
                            consol_gene,
                            pd.DataFrame(
                                [
                                    [
                                        bed_gene_data[transcript][0][0],
                                        longest_df[
                                            longest_df["consolidation_transcript_id"]
                                            == consol_transcript
                                        ]["consol_start"].values[0],
                                        longest_df[
                                            longest_df["consolidation_transcript_id"]
                                            == consol_transcript
                                        ]["consol_end"].values[0],
                                        bed_gene_data[transcript][0][5],
                                        gene_id,
                                        consol_transcript,
                                    ]
                                ],
                                columns=consol_gene.columns,
                            ),
                        ],
                        ignore_index=True,
                    )
            else:
                for row in sorted(bed_gene_data[transcript]):
                    num = num + 1
                    if num == 1:
                        consol_gene = pd.concat(
                            [
                                consol_gene,
                                pd.DataFrame(
                                    [
                                        [
                                            row[0],
                                            longest_df[
                                                longest_df["consolidation_transcript_id"]
                                                == consol_transcript
                                            ]["consol_start"].values[0],
                                            int(row[2]),
                                            row[5],
                                            gene_id,
                                            consol_transcript,
                                        ]
                                    ],
                                    columns=consol_gene.columns,
                                ),
                            ],
                            ignore_index=True,
                        )
                    elif num == len(bed_gene_data[transcript]):
                        consol_gene = pd.concat(
                            [
                                consol_gene,
                                pd.DataFrame(
                                    [
                                        [
                                            row[0],
                                            int(row[1]) + 1,
                                            longest_df[
                                                longest_df["consolidation_transcript_id"]
                                                == consol_transcript
                                            ]["consol_end"].values[0],
                                            row[5],
                                            gene_id,
                                            consol_transcript,
                                        ]
                                    ],
                                    columns=consol_gene.columns,
                                ),
                            ],
                            ignore_index=True,
                        )
                    else:
                        consol_gene = pd.concat(
                            [
                                consol_gene,
                                pd.DataFrame(
                                    [
                                        [
                                            row[0],
                                            int(row[1]) + 1,
                                            int(row[2]),
                                            row[5],
                                            gene_id,
                                            consol_transcript,
                                        ]
                                    ],
                                    columns=consol_gene.columns,
                                ),
                            ],
                            ignore_index=True,
                        )
                        consol_gene = consol_gene.sort_values(
                            by=["transcript_id", "start"]
                        ).reset_index(drop=True)
    logger.debug(
        "Gene {} has {} transcript(s) before consolidation and {} transcript(s) after "
        "consolidation.",
        gene_id,
        len(key_gene),
        key_gene["consolidation_transcript_id"].nunique(),
    )
    return consol_gene, key_gene


def consolidate_transcripts(data, consol_prefix, gene_list):
    """
    If requested, consolidate junctions in input transcripts.
    """
    # Loop over genes
    consol_data = pd.DataFrame(columns=data.columns)
    key = []

    for gene in gene_list:
        # Test for single transcript gene (WBGene00000003)
        # Test gene for multiple groups with consolidation (WBGene00001574)
        # Test monoexon gene (WBGene00000214)
        # Test monoexon with multiexon (WBGene00022161) -> added extra 2 monoexon transcripts
        # using gene_df =
        # pd.concat([gene_df,pd.DataFrame([["I",1779855,1781091,'-',"WBGene00022161","new_transcript_1"],["I",1781991,1782900,'-',"WBGene00022161","new_transcript_2"]],
        # columns=gene_df.columns)],ignore_index=True)
        pre_consol_jct = []
        gene_df = data[data["gene_id"] == gene]
        try:
            bed_gene_data = prep_bed_for_ea(gene_df)
        except ValueError as e:
            logger.error(e)
            continue
        transcripts = gene_df.groupby("transcript_id")
        # Loop over transcripts to get junctions
        for transcript in transcripts.groups:
            pre_consol_jct.extend(
                create_junction_catalog_str(gene, transcript, bed_gene_data[transcript])
            )
        pre_consol_jct_df = pd.DataFrame(pre_consol_jct, columns=jct_df_cols)
        # Consolidate 5'/3' variation
        consol_gene, key_gene = consolidate_junctions(
            bed_gene_data, pre_consol_jct_df, consol_prefix
        )
        consol_data = pd.concat([consol_data, consol_gene], ignore_index=True)
        key.append(key_gene)

    # Set data variable to new consolidated data
    data = consol_data.copy()
    del consol_data
    genes = data.groupby("gene_id")
    # Combine key file entries
    if len(key) > 1:
        keys = pd.concat(key, ignore_index=True)
    else:
        keys = key[0]
    # Add columns for number of transcript_id, number of consolidated_transcript_id
    #   and a flag for if the gene was consolidated
    keys["num_transcript_in_consol_transcript"] = keys.groupby(
            "consolidation_transcript_id")["transcript_id"].transform('nunique')
    keys["num_transcript_id_in_gene"] = keys.groupby(
            "gene_id")["transcript_id"].transform('nunique')
    keys["num_consol_transcript_id_in_gene"] = keys.groupby(
            "gene_id")["consolidation_transcript_id"].transform('nunique')
    keys["flag_gene_consolidated"] = np.where(
        keys["num_transcript_id_in_gene"] > keys["num_consol_transcript_id_in_gene"],
        1,
        0
    )

    return data, genes, keys



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jun 12 16:13:11 2021

@author: adalena.nanni
"""

import matplotlib.pyplot as plt

# Import plot functions
from . import plot_functions as PF


def plot_one_gtf_pairwise(outdir, td_data, prefix):
    """
    Plot 1 GTF Pairwise output including the following:

        1) Upset plot for the number of genes with each kind of alternative splicing (AS)

        2) Upset plot the number of transcript pairs with each kind of AS
           for a single transcriptome with added box plots
           of the number of nt different and the proportion of nt different
    """
# Plot 1
    # Upset plot for the number of genes with each kind of alternative splicing (AS)
    if prefix is not None:
        PF.plot_gene_AS_upset(td_data, "{}/{}_gene_AS_upset.rtf".format(outdir, prefix))
        plt.savefig("{}/{}_gene_AS_upset.png".format(outdir, prefix), dpi=600, format="png")
    else:
        PF.plot_gene_AS_upset(td_data, "{}/gene_AS_upset.rtf".format(outdir))
        plt.savefig("{}/gene_AS_upset.png".format(outdir), dpi=600, format="png")
    plt.clf()

# Plot 2
    # Upset plot the number of transcript pairs with each kind of
    #   AS for a single transcriptome with added box plots
    #   of the number of nt different and the proportion of nt different
    if prefix is not None:
        PF.plot_pair_AS_upset_nt_box(td_data, "{}/{}_transcript_pair_AS_upset_nt_boxplot.rtf".format(outdir, prefix))
        plt.savefig("{}/{}_transcript_pair_AS_upset_nt_boxplot.png".format(outdir, prefix), dpi=600, format="png")
    else:
        PF.plot_pair_AS_upset_nt_box(td_data, "{}/transcript_pair_AS_upset_nt_boxplot.rtf".format(outdir))
        plt.savefig("{}/transcript_pair_AS_upset_nt_boxplot.png".format(outdir), dpi=600, format="png")
    plt.clf()